/*------------------------------------------------------------------+
|Project    : Max Weighted Clique Branch and Bound with Bounded Size|      
|Copyright (c): 2017-2018 Daniel PORUMBEL, daniel.porumbel@cnam.fr  |       
|License: Creative Commons Attribution 3.0 Unported License         |
|         http://creativecommons.org/licenses/by/3.0/               |
+------------------------------------------------------------------*/
//TODO: remove useless OPTIM_USE_ONLY_ONE_SCAN, here and in .h

#include "cliqueBBBS.h"
#include <algorithm>         //for sort
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <climits>
#include <iostream>          //for the Dimacs graph reader
#include <fstream>           //for the Dimacs graph reader
#include <string.h>

using namespace std;
#define EPSILON 1.0e-6

extern int**   edge;           //These variables simply define the graph:
extern int     n;              //vtx no
extern int     m;              //edge no

namespace
{

double*  ww;                   //global vertex weights
int*   idx;                    //sorted indexes pointing to vertices, eg, ww[idx[0]]>=ww[idx[i]]
typedef struct _listCell{
    int v ;                    //selected vertex idx[v]
    struct _listCell * nxt;
}listCell;
listCell* clq       = NULL;

int szMax;                     //max cardinal/size
int szCurr;                    //szCurr current clique
double wCurr;                  //weight current clique
double lwBnd;                  //lwBnd to the best clique we know of (eg, given by heuristic or a clique discovered durring B&B
double wBst;                   //best found during B&B
int*   clqBst;                 //record here the best clique as an array

//when used with std::sort, it will sort in weight descendent order
bool myComp (int i,int j) { return (ww[i]>ww[j]); }

#ifdef USE_A_FAST_HEURISTIC_IN_BB
double heuristicInBB(int* clique, double*w, int szMax)
{
    int * deadidx = new int[n];
    int p=0;
    
    double sommeVal = 0;
    for(int i=0;i<(::n);i++){
        clique[i] = 0;
        deadidx[i] = 0;
    }
    deadidx[0] = 1;             //if you start at 0 you find the same solution as the B&B
    deadidx[3] = 1;         
    for(int k=1;k<=szMax;k++){
        for(;p<(::n);p++)
            if(!deadidx[p]){
                clique[idx[p]] = 1;
                sommeVal +=w[idx[p]];

                deadidx[p]=1;
                for(int q=p+1;q<(::n);q++)
                    if(!::edge[idx[q]][idx[p]])
                        deadidx[q] = 1;
                break;
            }
        if(p==(::n))
            break;
    }
    delete[] deadidx;
    return sommeVal;
}
#endif /*USE_A_FAST_HEURISTIC_IN_BB*/

#ifdef PRUNE_ON_NOT_USING_A_VTX
int**    nonEdgAdjLst;
double * maxWithout;
void nonEdgAdjLstFill(){
    ::maxWithout = new double[::n];
    ::nonEdgAdjLst    = new int*[::n];
    for(int i=0;i<n;i++){
        ::nonEdgAdjLst[i] = new int[n+1];
        int k = 0;
        for(int j=0;j<n;j++)
          if(i!=j)
            if(!::edge[j][i]){
                ::nonEdgAdjLst[i][k] = j;
                k++;
            }
        assert(k<=n);
        nonEdgAdjLst[i][k] = -1;
    }
    for(int i=0;i<n;i++)
        ::maxWithout[i] = 0;
}
void nonEdgAdjLstFree(){
    delete[] ::maxWithout;
    for(int i=0;i<n;i++)
        delete[] ::nonEdgAdjLst[i];
    delete[] nonEdgAdjLst;
}
#endif /* PRUNE_ON_NOT_USING_A_VTX */


inline int connectedToAllUpTo(int j, int upToSize)
{
    assert(upToSize>0);
    //assert(upToSize<=::szCurr);//it did find it true
    assert(clq->v>=0);
    assert(clq->v<(::n));
    assert(j>=0);
    assert(j<(::n));
    if(!::edge[idx[clq->v]][idx[j]])
        return 0;
    listCell* scanner = clq;
    for(int i=1;i<upToSize;i++){                //upToSize-1 nexts
        scanner = scanner->nxt;
        assert(scanner!=NULL);
        if(!::edge[idx[scanner->v]][idx[j]])
            return 0;
    }
    return 1;
}
//take current clique from clq and do szCurr-1 nexts
void clqBstWriteSol()
{
    listCell* scanner = clq;
    for(int i=0;i<(::n);i++)
        ::clqBst[i] = 0;
    ::clqBst[::idx[scanner->v]] = 1;
    for(int i=0;i<(::szCurr-1);i++){
        scanner = scanner->nxt;
        assert(scanner!=NULL);
        ::clqBst[::idx[scanner->v]] = 1;
    }
}
double upperBoundAtNode(int lstVtUsed)
{
    int j =lstVtUsed + 1;                          //vertices we could add are called j
    assert(j>=0);
    double upBound = ::wCurr;
    #ifdef BETTER_UB
    int maxShadowingVtx = MAX_SHADOWING_VTX;//MAX_SHADOWING_VTX can do calculations, maxShadowingVtx no
    double avgDeg = ((double)::m)/(::n);
    if(szMax*2>avgDeg){                 //special strange graph
        //cout<<"The clique size is more than half the average degree"
        //    <<"I won't try faster upper bounds in clique BB\n";
        maxShadowingVtx = 0;
    }
    maxShadowingVtx = min(::szMax-::szCurr-1,maxShadowingVtx);   //inClq has to be useful for at least a vertex, hence -1
    int*  inClq = new int[maxShadowingVtx];
    #ifdef EVEN_BETTER_UB_USING_NON_TRIANGLES
    int*  nonTri= new int[maxShadowingVtx];
    #endif
    int szNewClqSeg = 0;
    for(int i=::szCurr+1;i<=(::szMax);i++){
        int vtxConnected = -1;
        for(;j<(::n);j++){
            if(connectedToAllUpTo(j,szCurr)){
                  int vtxUseful = 1;
                  for(int p=0;p<szNewClqSeg;p++){
                      if(inClq[p]>=0){
                          if(!::edge[idx[inClq[p]]][idx[j]]){
                              vtxUseful = 0;
                              #ifdef EVEN_BETTER_UB_USING_NON_TRIANGLES
                              inClq[p]  = inClq[p]-10000; //disable 
                              nonTri[p] = j;
                              #else
                              inClq[p]  = -1;       //disable 
                              #endif
                              break;
                          }
                      }
                      #ifdef EVEN_BETTER_UB_USING_NON_TRIANGLES
                      else{
                        if(inClq[p]>=-10000){
                          if(!::edge[idx[inClq[p]+10000]][idx[j]])
                            if(!::edge[idx[nonTri[p]]][idx[j]]){
                                vtxUseful = 0;
                                inClq[p]=-1000000;
                            }
                        }
                      }  
                      #endif /*EVEN_BETTER_UB_USING_NON_TRIANGLES*/
                  }
                  if(vtxUseful){
                      vtxConnected = j;
                      upBound    += ww[idx[j]];
                      //Below seems useless, although legitimate
                      //if(upBound>::lwBnd) 
                      //       goto out_free_mem;
                      if(szNewClqSeg<maxShadowingVtx){
                           inClq[szNewClqSeg] = j;
                           szNewClqSeg++;
                      }
                      j++;
                      break;    //go to next i
                  }
            }//endif found connected j
        }
        //if(vtxConnected>0) assert(j==vtxConnected+1);
        if(vtxConnected==-1)                          //no more vtx that can form a clique
            goto out_free_mem;                        //goto ok in this case for the Linux kernel coding style
    }
    out_free_mem:
    #ifdef EVEN_BETTER_UB_USING_NON_TRIANGLES
    delete[] nonTri;
    #endif
    delete[] inClq;
    return upBound;
    #else
    for(int i=::szCurr+1;i<=(::szMax);i++){
        int vtxConnected = -1;
        for(;j<(::n);j++){
            if(connectedToAllUpTo(j,szCurr)){
              vtxConnected = j;
              upBound    += ww[idx[j]];
              j++;
              break;
            }
        }
        //if(vtxConnected>0) assert(j==vtxConnected+1);
        if(vtxConnected==-1)                          //no more vtx that
            return upBound;                           //can form a clique
    }
    return upBound;
    #endif
}

#ifdef OPTIM_USE_ONLY_ONE_SCAN
#ifdef BETTER_UB
//consider current clique from clq on which we can do szMax-szCurr nexts
double upperBoundAtNode(listCell*clqConn)
{
    double upBound = ::wCurr;
    //old version of better ub, not so good
    //first
    if(clqConn==NULL)
        return upBound;
    int fstVtxConnected = clqConn->v;
    upBound    += ww[idx[fstVtxConnected]];
    //second
    clqConn = clqConn->nxt;
    if(clqConn==NULL)
        return upBound;
    int secVtxConnected = clqConn->v;
    
    if(edge[::idx[secVtxConnected]][::idx[fstVtxConnected]]){
            upBound    += ww[idx[secVtxConnected]];
            clqConn = clqConn->nxt;
            while(clqConn!=NULL){
                upBound    += ww[idx[clqConn->v]];
                clqConn = clqConn->nxt;
            }
            return upBound;
    }
    //from third on
    int lastAdded = clqConn->v;
    clqConn = clqConn->nxt;
    while(clqConn!=NULL){
        upBound    += ww[idx[clqConn->v]];
        lastAdded = clqConn->v;
        clqConn = clqConn->nxt;
    }
    for(int j=lastAdded+1;j<(::n);j++)
        if(connectedToAllUpTo(j,szCurr))
            return upBound+ww[idx[j]];

    return upBound;
}
#else  /* !BETTER_UB */
//consider current clique from clq on which we can do szMax-szCurr nexts
double upperBoundAtNode(listCell*clqConn)
{
    double upBound = ::wCurr;
    while(clqConn!=NULL){
        upBound    += ww[idx[clqConn->v]];
        clqConn = clqConn->nxt;
    }
    return upBound;
}
#endif /* BETTER_UB */
int findClqConn(int lstVtConn, listCell* &clqConn, int howManyToSearch)
{
    for(int j=lstVtConn+1;j<(::n);j++)
        if(connectedToAllUpTo(j,szCurr)){
            clqConn = new listCell;
            clqConn-> v = j;
            clqConn-> nxt = NULL;
            if(howManyToSearch==1)
                return j;
            return findClqConn(j,clqConn->nxt,howManyToSearch-1);
        }
    return ::n;
}
#endif /*OPTIM_USE_ONLY_ONE_SCAN*/
//suppose curr is allocated and it is safe to write on it values from lstVtUsed+1 to ::n-1
void nodeBB(int lstVtUsed, listCell*curr)
{
    if(::wCurr  >  ::wBst){
        ::wBst     = ::wCurr;
        ::lwBnd    = max(::lwBnd,::wBst);
        clqBstWriteSol();
    }
    if(::szCurr+1  >  ::szMax)
        return ;
    #ifdef OPTIM_USE_ONLY_ONE_SCAN
    listCell* clqConn = NULL;                     //vertices connected to curr Clique
    int lstCliqueConn = lstVtUsed;
    if(::szMax-::szCurr>3){
        lstCliqueConn = findClqConn(lstVtUsed, clqConn,::szMax-::szCurr);
        if(upperBoundAtNode(clqConn)<=::lwBnd+EPSILON){
            while(clqConn!=NULL){
                    listCell* deleter = clqConn;
                    clqConn = clqConn-> nxt;
                    delete deleter;
            }
            return;
        }
    }else{
        if(upperBoundAtNode(lstVtUsed)<=::lwBnd+EPSILON)
            return ;
    }
    //assert(upperBoundAtNode(lstVtUsed)==upperBoundAtNode(clqConn));
    #else
    if(upperBoundAtNode(lstVtUsed)<=::lwBnd+EPSILON){
        //for(int ii=0;ii<szCurr;ii++)
        //    cout<<" ";
        //cout<<"at "<<szCurr<<":CUTbecause"<<endl;
        return ;
    }
    #endif

    #ifdef PRUNE_ON_NOT_USING_A_VTX
    int* nonEdges ;
    nonEdges = nonEdgAdjLst[idx[lstVtUsed]];
    for(int k=0;nonEdges[k]!=-1;k++){
        int jj          = nonEdges[k];
        maxWithout[jj] += ::ww[idx[lstVtUsed]];
    }
    maxWithout[idx[lstVtUsed]] += ww[idx[lstVtUsed]]+1;
    //check if we can prune
    int elemsMaxToAdd = ::szMax-::szCurr;     //consider lstVtxUsed is already added
    for(int iii=0;iii<lstVtUsed;iii++){
        int ii = idx[iii];
        double maxWithout_ii = maxWithout[ii];
        int addedElems       = 0;
        for(int j=lstVtUsed+1;j<(::n);j++){
            if(maxWithout_ii>(::ww[ii]))
                break;
            if(addedElems>=elemsMaxToAdd)
                break;
            if(!::edge[ii][idx[j]])
            if(connectedToAllUpTo(j,szCurr))
            {
                maxWithout_ii+=::ww[idx[j]]; 
                addedElems++;
            }
        }
        if(maxWithout_ii<=(::ww[ii])){
            //listCell* scanner = clq;
            //    cout<<scanner->v<<::edge[idx[scanner->v]][idx[ii]]<<endl;
            //for(int i=1;i<szCurr;i++){                //upToSize-1 nexts
            //    scanner = scanner->nxt;
            //    cout<<scanner->v<<::edge[idx[scanner->v]][idx[ii]]<<endl;
            //}
            //        cout<<"wCurr="<<wCurr<<endl;
            //        cout<<"p"<<":"<<lstVtUsed<<"z"<<ii<<" "<<maxWithout_ii<<" "<<addedElems<<","<<elemsMaxToAdd<<"|"<<flush;
            //        if(::wCurr!=64000)
            if(addedElems>0){
                //cout<<"p";
                goto remove_lst_contrib_exit;
            }
        }
    }
    #endif
    curr->nxt = new listCell;                   //prepare the cell where next nodeBB can safely write
    ::szCurr ++;
    //clog<<"szCurr="<<szCurr<<endl;
    #ifdef OPTIM_USE_ONLY_ONE_SCAN
    while(clqConn!=NULL){
            int j = clqConn->v;
            curr->v = j;
            ::wCurr += ww[idx[j]];
            nodeBB(j,curr->nxt);
            ::wCurr -= ww[idx[j]];
            listCell* deleter = clqConn;
            clqConn = clqConn-> nxt;
            delete deleter;
    }
    for(int j=lstCliqueConn+1;j<(::n);j++)
        if(connectedToAllUpTo(j,szCurr-1)){         //at this moment szCurr includes the potential new vtx j
            curr->v = j;
            ::wCurr += ww[idx[j]];
            nodeBB(j,curr->nxt);
            ::wCurr -= ww[idx[j]];
        }
    #else
    for(int j=lstVtUsed+1;j<(::n);j++){
        if(connectedToAllUpTo(j,szCurr-1)){         //at this moment szCurr includes the potential new vtx j
            curr->v  = j;
            ::wCurr += ww[idx[j]];
            nodeBB(j,curr->nxt);
            ::wCurr -= ww[idx[j]];
        }
    }
    #endif
    szCurr--;
    //clog<<"szCurr="<<szCurr<<endl;
    delete curr->nxt;
    #ifdef PRUNE_ON_NOT_USING_A_VTX
    remove_lst_contrib_exit:
    maxWithout[idx[lstVtUsed]] -= (ww[idx[lstVtUsed]]+1);
    for(int k=0;nonEdges[k]!=-1;k++){
        int jj          = nonEdges[k];
        maxWithout[jj] -= ::ww[idx[lstVtUsed]];
    }
    #endif
}

}//end unnamed namespace for avoiding name collisions

//return best clique width if size of szMax found or -missed edges otherwise
double fastHeuristicBS(int* clique, double*w, int szMax)
{
    ::ww = w;
    ::idx      = new int[::n];
    for(int i=0;i<(::n);i++)
        ::idx[i]=i;
    ::sort(idx, idx+::n,myComp);                                

    int * missed_edg_idx = new int[n];
    int p=0;
    int k;
    
    double sommeVal = 0;
    for(int i=0;i<(::n);i++){
        clique[i] = 0;
        missed_edg_idx[i] = 0;
    }
    for(k=1;k<=szMax;k++){
        for(;p<(::n);p++)
            if(missed_edg_idx[p]==0){
                clique[idx[p]] = 1;
                sommeVal +=w[idx[p]];

                missed_edg_idx[p]=1;
                for(int q=0;q<(::n);q++)
                    if(!::edge[idx[q]][idx[p]])
                        missed_edg_idx[q]++;
                break;
            }
        if(p==(::n)){//failed to find szMax-clique
            sommeVal = 0;
            int missedMin ;
            int bst;
            for(;k<=szMax;k++){
                missedMin = INT_MAX;
                bst       = -1;
                for(p=0;p<(::n);p++){
                    if(!clique[idx[p]]){
                        if(missed_edg_idx[p]<missedMin){
                            missedMin = missed_edg_idx[p];
                            bst       = p;
                        }
                    }
                }
                assert(bst>=0);
                clique[idx[bst]]= 1;
                sommeVal+=missedMin;
                for(int q=0;q<(::n);q++)
                    if(!::edge[idx[q]][idx[bst]])
                        missed_edg_idx[q]++;
            }
            sommeVal = -sommeVal;
            goto cleanup_return;
        }
    }
    cleanup_return:
    delete[] missed_edg_idx;
    delete[] idx;
    return sommeVal;
}
//return best clique value. Since we know there is a clique of weigth heuristicLb, the return 
//value is >= heuristicLb. However, it fills the clique vector only if it finds 
//a strictly better clique than heuristicLb. We suppose we do have a clique for heuristicLb
//do not search cliques with more than maxCardinal elements
double maxWeightCliqueBBBS(int * clique, double * w, double heuristicLb, int maxCardinal)
{
    ::ww       = w;
    ::wBst     = INT_MIN;                         //weight of best clique found
    ::wBst     = heuristicLb;   
    ::lwBnd    = heuristicLb;
    ::wCurr    = 0;                               //weight current clique
    ::szCurr   = 0;                               //size current clique
    ::szMax    = maxCardinal;
    ::clq      = new listCell;
    ::clq->nxt = new listCell;                    //next will be given to nodeBB as safe place to write vertices after i
    ::clqBst   = clique;
    ::idx      = new int[::n];
    for(int i=0;i<(::n);i++)
        ::idx[i]=i;
    ::sort(idx, idx+::n,myComp);                                
    #ifdef USE_A_FAST_HEURISTIC_IN_BB
    if(heuristicLb<=0) {
        heuristicLb = heuristicInBB(clique,w,maxCardinal);
    }
    #endif
    #ifdef PRUNE_ON_NOT_USING_A_VTX
    nonEdgAdjLstFill();
    #endif
    for(int i=0;i<(::n-1);i++)
        assert(ww[idx[i]]>=ww[idx[i+1]]);
    for(int i=0;i<(::n);i++){
        ::clq->v = i;
        ::szCurr = 1;
        ::wCurr = ww[idx[i]];
        nodeBB(i, ::clq->nxt);
    }
    delete[] clq->nxt;
    delete[] clq;
    delete[] ::idx;
    //You might increase the clique size by adding more vtx with code below.
    /*
    int * touched = new int[::n];
    int clqSizeNow = 0;
    for(int i=0;i<(::n);i++){
        touched[i] = clique[i];
        clqSizeNow++;
    }
    while(clqSizeNow<maxCardinal){
        for(int i=0;i<(::n);i++)
            for(int j=0;j<(::n);j++)
                if(::edge[i][j]){
                   if(clique[i])
                        touched[j] = 1; 
                   if(clique[j])
                        touched[i] = 1; 
                }
        //int noWeightVtxAdded = 0;
        for(int i=0;i<(::n);i++)
            if(!touched[i]){
                assert(w[i]<EPSILON);
                clique[i] = 1;
                cerr<<"Finaly"<<endl;
                clqSizeNow++;
                exit(1);
        //        noWeightVtxAdded++;
            }
        //cout<<"noWeightVtxAdded="<<noWeightVtxAdded<<endl;
    }
    //*/
    //cout<<::wBst<<","<<flush;
    #ifdef PRUNE_ON_NOT_USING_A_VTX
    nonEdgAdjLstFree();
    #endif
    return ::wBst;
}
double maxWeightStableBBBS(int * clique, double * w, double heuristicLb, int maxCardinal)
{
    for(int i=0;i<(::n);i++)
        for(int j=0;j<i;j++){
            ::edge[i][j] = 1-::edge[i][j];
            ::edge[j][i] = 1-::edge[j][i];
        }
    double resultat = maxWeightCliqueBBBS(clique,w,heuristicLb,maxCardinal);
    for(int i=0;i<(::n);i++)
        for(int j=0;j<i;j++){
            ::edge[i][j] = 1-::edge[i][j];
            ::edge[j][i] = 1-::edge[j][i];
        }
    return resultat;
}
double maxWeightCliqueBBBS(int * clique, double * w, int maxCardinal)
{
    return maxWeightCliqueBBBS(clique, w, 0, maxCardinal);
}
double maxWeightStableBBBS(int * clique, double * w, int maxCardinal)
{
    return maxWeightStableBBBS(clique, w, 0, maxCardinal);
}



void readGraphDimacs(char*filename)
{
     ifstream inFile;
     inFile.open(filename);
     if ( inFile.fail() ){
           cerr << " Error opening " << filename << endl;
           exit(0);
     }
     char token[100];
     inFile >> token;
     if (inFile.eof()){
           cerr << " Found EOF while reading in the beginning " << filename << endl;
           exit(0);
     }
     int x1, x2;     
     int mCount=0;
     while ( ! inFile.eof() ) {
        if (!strcmp(token, "p" )) {
            char dummyWord[50];
            inFile >> dummyWord >> ::n >> ::m;
            clog << "Vertices = " << ::n << ",";
            clog << "Edges = " << ::m << "\n";               
            ::edge=new int*[::n];
            for(int ii=0;ii<(::n);ii++){
                ::edge[ii]=new int[::n];
                for(int jj=0;jj<(::n);jj++)
                    ::edge[ii][jj]=0;
            }
        }                                  
        if (!strcmp(token, "e")){
            inFile >> x1 >> x2;
            x1--; x2--;
            if ( x1<0 || x2<0 || x1>=n || x2 >=n ){
                  cerr << "Error reading ::edges x1=" << x1 << ", x2=" << x2 << endl;
                  exit(0);
            }
            ::edge[x1][x2]=1;
            ::edge[x2][x1]=1;
            mCount++;
        }
        inFile >> token;
     }  
     if ( (::m != mCount)&&(::m!=mCount*2))
           cerr<<"Warning reading graph. Announced m size=" << m << ", found=" << mCount  << endl;
     inFile.close();   
};

