/*---------------------------------------------------------------------------------------+
|                                                                                        |
|                                     Clique-BBBS                                        |
|        Max Weighted Clique Branch and Bound with Bounded Size (max nr of elems)        |
|                                                                                        |
|    I wrote this code because I needed to solve a weighted maximum clique problem with  |
|  bounded maximum cardinal (nb of elems) and I could not find a program elsewhere. This |
|  solution seems much faster than the one obtained my modelling the problem in Cplex.   |
|  This code is distributed in the hope that it will be useful, but without any warranty.|
|                                                                                        |
-----------------+---------------------------------------------------------------+-------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
/*see accompanying .cpp file for licence info   */

#ifndef CLIQUEBBBS_H_INCLUDED
#define CLIQUEBBBS_H_INCLUDED

#define BETTER_UB                           //These two macros can bring an important 
#define EVEN_BETTER_UB_USING_NON_TRIANGLES  //speed-up (7x in the below example) 
                                            //These macros are only used if szMax < avg (deg)/2

#define MAX_SHADOWING_VTX (min(::szMax*2/3,10)) //how many shadowing vertices might be considered

/* Basic idea of better UB: consider clique [1..s] recalling w[1]>=w[2]>=....w[n]
    say s1 is the next that could be in a clique: add it to ub
    say s2 is the next that could be in a clique: add it to ub 
    say s3 is the next that could be in a clique but {s1,s3} is not an edge:
                  : do not add s3, s1 shadows s3!! Only one of them can contribute to UB
    say s4 is the next that could be in a clique but {s1,s4} is not and edge:
                  : add s4, s1 can not shadow s4 as well, never try to make s1 shadow 
                    vertices this way (selecting s3 and s4 could be better than s1)
    say s5 is ok, add s5. If SZ_CLQ_ADDED is reached, do not try to make s5 shadow vertices

   When using NON_TRIANGLES, after s1 shadows s3, they are put in a list of
    non-edges. If {s1, s3, s6} is stable, s1 shadows both s3 and s6. As for the
    s4 example above, such shadowing can happen only once. TODO: NON_TRIANGLES->3STABLES
   This is described in Appendix A.3 of the paper.
*/

//#define USE_A_FAST_HEURISTIC              // It brings nothing decisive, if the provided lower bound is zero, uses a heuristic to get a lower bound. 

//#define OPTIM_USE_ONLY_ONE_SCAN           //not useful, another way of scanning vtx connected to curr Clique

//#define PRUNE_ON_NOT_USING_A_VTX          // Use a pruning technique from Section 2.1.1. (line 5, Alg 1) of [Held et al, 
                                            // Maximum-Weight Stable Sets and Safe Lower Bounds For Graph Coloring, Math Prog Comp]


//return best clique value. Assuming we already have a clique of weight heuristicLb, the return 
//value is >= heuristicLb. However, it fills the clique vector only if it finds 
//a strictly better clique than heuristicLb with no more than maxCardinal elements
double maxWeightCliqueBBBS(int * clique, double * w, double heuristicLb, int maxCardinal);
double maxWeightStableBBBS(int * clique, double * w, double heuristicLb, int maxCardinal);

//Versions of above functions that simply set heuristicLb=0 
double maxWeightCliqueBBBS(int * clique, double * w, int maxCardinal);
double maxWeightStableBBBS(int * clique, double * w, int maxCardinal);

//return best clique width if size of szMax found or -missed edges otherwise
double fastHeuristicBS(int* clique, double*w, int maxCardinal);

//This will fill the global data structures (solved by above functions) with the graph from filename.
void readGraphDimacs(char*filename);


#endif

/*//This is a usage example (while it compiles here, better put it in the cpp file)
#include <cstdlib>
#include <iostream>          //for the Dimacs graph reader
using namespace std;

int**   edge;           //These variables simply define the graph:
int     n;              //vtx no
int     m;              //edge no
int main(){
    n = 400; 
    m = 70000; 
    edge = new int*[n];
    for(int i=0;i<n;i++){
        edge[i] = new int[n];    
        for(int i2=0;i2<n;i2++)
            edge[i][i2] = 0;
    }
    cout<<"Generating graph with density="<<((double)m*2)/((double)(n*n-1))<<endl;
    for(int j=0;j<m;j++){
        int v1,v2;
        int triesGen = 0;
        do{
            v1 = rand()%n;
            v2 = rand()%n;
            triesGen ++;
        }while( (edge[v1][v2])&&(triesGen<n));
        if(edge[v1][v2]){                     //too many failed tries
            cerr<<"I think I can not generate "<<m<<" edges for a graph of order "<<n<<endl;
            exit(EXIT_FAILURE);
        }
        edge[v1][v2] = 1;
        edge[v2][v1] = 1;
    }
    //calling max clique algorithm
    int*    solution = new int[n];
    double* weights  = new double[n];
    double heuristSol= 0;
    for(int i=0;i<n;i++){
        double prob  = (double)(rand()%1000)/500;
        weights[i]   = 1000*prob*prob;
        //cout<<weights[i]<<endl;
    }
    int maxVtxInClq  = n/15;
    cout<<"Searching max weighted clique with cardinal="<<maxVtxInClq<<"..."
        <<"(this might take 10-15 secs)..."<<flush;
    cout<<"->"<<maxWeightCliqueBBBS(solution, weights, heuristSol, maxVtxInClq)<<endl;
    //Analyse response
    int cardClq = 0;
    for(int i=0;i<n;i++)
        cardClq+=solution[i];
    cout<<"Cardinal clique obtained="<<cardClq<<endl;
}
*/
