/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include <iostream>
#include <vector>
#include <list>
#include <cstdlib>
#include <cassert>
#include <string>
#include <sstream>
 
#include <limits> // for numeric_limits
#include <climits> // for numeric_limits
 
#include <set>
#include <utility> // for pair
#include <algorithm>
#include <iterator>

#include "dijkstra.h"

 
namespace{ 
   struct neighbor {
       int target;
       weight_t weight;
       neighbor(int arg_target, weight_t arg_weight)
           : target(arg_target), weight(arg_weight) { }
   };
    
   int nbVtx;
   typedef vector<vector<neighbor> > adj_list_t;
   adj_list_t adj_list;
   vector<weight_t> minDist;
   vector<int>      prevRel;
}//end namespace of dijkstra cpp

void dijkstraCalcPaths(int src, weight_t* distances){
    //erase previous data if any
    vector<weight_t>().swap(::minDist);
    vector<int>().swap(::prevRel);
    dijkstraCalcPaths(src,::minDist,::prevRel);
    for(int i=0;i<(::nbVtx);i++)
        distances[i] = ::minDist[i];
}
void getVtxOnPathTo(int dst, int* vertices){
    for(int i=0;i<(::nbVtx);i++)
        vertices[i] = 0;
    int currVtx = dst;
    vertices[currVtx] = 1;
    for ( ; currVtx != -1; currVtx = ::prevRel[currVtx])
        vertices[currVtx]  = 1;
}
 
void dijkstraFreeMem(){
    for(int i=0;i<(::nbVtx);i++)
        vector<neighbor>().swap(adj_list[i]);//free memory
    vector<vector<neighbor> >().swap(adj_list);
    vector<weight_t>().swap(minDist);
    vector<int>().swap(::prevRel);
}

void dijkstraCalcPaths(int source,
                          vector<weight_t> &min_dist,
                          vector<int> &previous)
{
    int n = nbVtx;
    min_dist.clear();
    min_dist.resize(n, INT_MAX);
    min_dist[source] = 0;
    previous.clear();
    previous.resize(n, -1);
    std::set<std::pair<weight_t, int> > vertex_queue;
    vertex_queue.insert(std::make_pair(min_dist[source], source));
 
    while (!vertex_queue.empty()) 
    {
        weight_t dist = vertex_queue.begin()->first;
        int u = vertex_queue.begin()->second;
        vertex_queue.erase(vertex_queue.begin());
 
        // Visit each edge exiting u
	const std::vector<neighbor> &neighbors = adj_list[u];
        for (std::vector<neighbor>::const_iterator neighbor_iter = neighbors.begin();
             neighbor_iter != neighbors.end();
             neighbor_iter++)
        {
            int v = neighbor_iter->target;
            weight_t weight = neighbor_iter->weight;
            weight_t dist_through_u = dist + weight;
	    if (dist_through_u < min_dist[v]) {
	        vertex_queue.erase(std::make_pair(min_dist[v], v));
 
	        min_dist[v] = dist_through_u;
	        previous[v] = u;
	        vertex_queue.insert(std::make_pair(min_dist[v], v));
 
	    }
 
        }
    }
}
 
 
std::list<int> dijkstraGetShPathTo( int vertex, const std::vector<int> &previous) {
    std::list<int> path;
    for ( ; vertex != -1; vertex = previous[vertex])
        path.push_front(vertex);
    return path;
}

string dijkstraGetStrShPathTo( int vertex, const std::vector<int> &previous) {
    
    list<int> path = dijkstraGetShPathTo(vertex, previous);

    ostringstream ss;
    ss<<path.front();
    for (path.pop_front();!path.empty();path.pop_front()){
         ss<<"->";
         ss<<path.front();
    }
    ss<<"\n";
    return ss.str();
}
 

void setGraphSize(int sz){
    nbVtx = sz;
    adj_list.resize(sz); 
}
void addEdge(int i, int j, weight_t weight){
    assert(weight>=0);
    adj_list[i].push_back(neighbor(j, weight));
    adj_list[j].push_back(neighbor(i, weight));
} 
void removeLastVt(){
    for(int i=0;i<nbVtx-1;i++){
        if(adj_list[i].back().target==nbVtx-1){
            adj_list[nbVtx-1].pop_back();
        }
    }
    adj_list[nbVtx-1].clear();
}
double linkLastVtx(weight_t * newLinksLastVtx){
    //clog<<endl;
    double offset = INT_MAX;
    for(int i=0;i<nbVtx-1;i++)
        if(newLinksLastVtx[i]>INT_MIN){
          if(-newLinksLastVtx[i]<offset){
                offset = -newLinksLastVtx[i];
            }
          //clog<<"["<<nbVtx-1<<","<<i<<"]:"<<newLinksLastVtx[i]<<",";
        }
    //clog<<endl;
    
    for(int i=0;i<nbVtx-1;i++)
        if(newLinksLastVtx[i]>INT_MIN){
            addEdge(nbVtx-1,i,-newLinksLastVtx[i]-offset);
        }
    return offset;
}


