/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include "graphstables.h"
#include "cliqueBBBS.h"
#include <iostream>
#include <algorithm>         //for sort
#include <fstream>
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <string.h>
#include <ilcplex/ilocplex.h>                    

#define ITER_LOC_SEARCH (::n*5)

using namespace std;

/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                                CPLEX weighted stable finder                                      |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/
                            
//The number of threads to use in cplex solver. Let 0 the default and it will choose 
//the nr of CPUs or (hyper-threads)
#define THREADS 0
//this file should be stand-alone, so we define a specific epsilon
#define EPSILON 1.0e-6

//call Cplex to find the max weighted (integer) stable
double maxWeightStableCplex(int* stable, double * w, int binary, int solveStabMax, int timeout)
{
    IloEnv         env;                          
    IloModel       model(env);
    IloRangeArray  constraints(env);                     
    IloNumVarArray vars(env,::n,0,1);
    IloCplex       cplex(env) ;
    int            nbConstraints = 0;
    cplex.setParam(IloCplex::Threads, THREADS);
    if(binary){
        model.add(IloConversion(env, vars, ILOINT));
        //the first below seems an old version (<Cplex 2.6) of second below
        cplex.setParam(IloCplex::EpInt, EPSILON);
        #ifndef CPLEXVER
           cplex.setParam(IloCplex::Param::MIP::Tolerances::Integrality,EPSILON);//comment this for cplex <12.6 and maybe also use SYSTEM=x86-64_sles10_4.1 in Makefile
        #else
           //CPLEXVER passed in command line via -D. It should contain 3 digits
           #if (CPLEXVER>=126)
               cplex.setParam(IloCplex::Param::MIP::Tolerances::Integrality,EPSILON);//you said cplex >=12.6
           #endif
        #endif

    }
    #ifdef NO_CPLEX_OUTPUT
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setWarning(env.getNullStream());
    #endif 
    //Not sure below is necessary
    cplex.setParam(IloCplex::NumericalEmphasis, true);

    if(solveStabMax)                            //do not put limit when searching cliques
    if(timeout>0){
        cplex.setParam(IloCplex::ClockType,1);  //timeout expressed in CPU time
        cplex.setParam(IloCplex::TiLim,timeout);
    }

    //objective
    IloExpr expr(env);
    for(int i=0;i<(::n);i++)
            expr+= w[i]*vars[i];
    IloObjective obj(env, expr, IloObjective::Maximize);
    model.add(obj);
    expr.end();
    //constraints, one per edge as in standard max clique/stable
    for(int i=0;i<(::n);i++)
        for(int j=0;j<i;j++)
            if( ((solveStabMax)&&(edge[i][j]))||    //search stables
                ((!solveStabMax)&&(!edge[i][j])))   //search cliques
            {
                IloExpr exprConstr(env);
                exprConstr=vars[i]+vars[j];
                constraints.add(exprConstr<=1);
                nbConstraints++;
                model.add(constraints[nbConstraints-1]);
                exprConstr.end();    
            }
    //constraints one per vertex for defective graph coloring with d=3
    //A vtx can have d neighbors max of the same color (d=0 => std coloring)
    //for(int i=0;i<(::n);i++){
    //    IloExpr exprConstr(env);
    //    exprConstr = ::n*vars[i]-::n;
    //    for(int j=0;j<i;j++)
    //        if( ((solveStabMax)&&(edge[i][j]))||    //search stables
    //            ((!solveStabMax)&&(!edge[i][j])))   //search cliques
    //        {
    //            exprConstr+=vars[j];
    //        }
    //    constraints.add(exprConstr<=3);             //max d=3 neighbors of same color
    //    nbConstraints++;
    //    model.add(constraints[nbConstraints-1]);
    //    exprConstr.end();    
    //}

    if(binary>1){
        IloExpr exprSum(env);
        for(int i=0;i<(::n);i++)
            exprSum+=vars[i];
        model.add(exprSum<=binary);
        exprSum.end();    
    }
    //(extract and) solve
    cplex.extract(model);
    try{
        cplex.solve();
    }catch (IloCplex::Exception e){
        cerr<<"Exception"<<e<<endl;
        if(cplex.getStatus()==IloAlgorithm::InfeasibleOrUnbounded){
            cerr<<"Attention: InfeasibleOrUnbounded. Strange, as all vars are <=1\n";
        }
        cerr<<"Saving the model to trycatch.lp. Check it\n";
        cplex.exportModel("trycatch.lp");
    }

    double maxWStable      = cplex.getObjValue();
    double maxWStableCheck = 0;
    //retrieve the stable

    IloNumArray solution(env);
    if(binary){
        cplex.getValues (vars,solution);
        for(int i=0;i<n;i++){
                //normal approach below leads to problems, solution[i]=1 and stable[i]=0
                //use below if you don't set NumericalEmphasis true
                if(solution[i]>0.5)
                    stable[i] = 1;
                else
                    stable[i] = 0;
                assert(stable[i]>=-EPSILON);
                maxWStableCheck+=stable[i]*w[i];
        }
        assert(abs(maxWStable-maxWStableCheck)<EPSILON);
    }
    #ifndef NDEBUG
    if(binary)
    for(int i=0;i<(::n);i++)
       for(int j=i+1;j<(::n);j++)
            if( ((solveStabMax)&&(edge[i][j]))||    //search stables
                ((!solveStabMax)&&(!edge[i][j])))   //search cliques
            {
                if((stable[i])&&(stable[j])){
                    cerr<<"Stable Cplex solver not good. It put "<<i<<" and "<<j<<"in \
                           the stable, but they can not belong together\n";
                }
            } 
    #endif
    if(cplex.getCplexStatus() ==IloCplex::AbortTimeLim){
        cerr<<"Attention: max weighted stable not solved to optimality, stopped after "<<timeout<<" seconds.\n";
        cerr<<"Suboptimal stablemax="<<maxWStable<<endl;
        maxWStable = -maxWStable;
    }

    /*Cleaning data and exit*/
    //Not sure it is necessary to end constraints one by one
    //cerr<<"Cleaning "<<nbConstraints<<" constraints\n";
    //for(int i=nbConstraints-1;i>=0;i--)
    //    constraints[i].end();
    constraints.end();
    vars.end();
    model.end();
    env.end();
    if((maxWStable<=0)&&(maxWStable>=-1)){ //the non-optimal stable does not cut current duals
        cerr<<"Doubling timeout because the non-optimal does not cut current duals\n";
        timeout*=2;
        if(timeout<100000)
            return maxWeightStableCplex(stable, w, binary, timeout);
    }
    return maxWStable;
}
double maxWeightStableCplex(int* stable, double * w, int binary, int timeout){
    return maxWeightStableCplex(stable,w,binary,1, timeout);
}
double maxWeightCliqueCplex(int* stable, double * w, int binary, int timeout){
    return maxWeightStableCplex(stable,w,binary,0,timeout);
}


/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                                         My local search                                          |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/
//Runs a local search to find a maximum clique starting from input clique
//Consider ::edge weights ww, some of them may be very negative (missing ::edge)
//Write the best found clique in clique, only if better than bstClqExisting
//Attention: you need to have an existing clique in clique (starting solution)
double mainLSEngine(int* clique, double** ww, int maxIter, double bstClqExisting)
{
    int* currClq    = new int[::n];
    double   bestObj= INT_MIN;
    int*       tabu = new int[::n];
    double* contrib = new double[::n];  //the (potential contribution to the obj funct
    double obj      = 0;
    for(int i=0;i<n;i++){
        tabu[i]   = 0;
        currClq[i] = 0;
        contrib[i]= 0;
    }

    for(int i=0;i<(::n);i++)
        currClq[i] = clique[i];

    for(int i=0;i<(::n);i++)
        if(currClq[i]){
            for(int j=i+1;j<(::n);j++)
                if(currClq[j])
                   obj+=ww[i][j];
            for(int j=0;j<(::n);j++)
                contrib[j] +=ww[i][j];
        }
    for(int iter=1;iter<=maxIter;iter++){
        //clog<<"Ls iter="<<iter<<" obj="<<obj<<endl;
        //1. choose non tabu move: select out deselect in; 
        int out =-1, in=-1;
        double bestContrib = INT_MIN;
        for(int i=0;i<(::n);i++)
            if((currClq[i])&&(tabu[i]<iter))
                for(int j=0;j<(::n);j++)
                    if((!currClq[j])&&(tabu[j]<iter)){
                        double contribNow = contrib[j]-contrib[i]-ww[i][j];
                        if(contribNow>bestContrib){ 
                            bestContrib = contribNow;
                            in  = i;
                            out = j;
                        }
                    }
        if(in<0)                 //everything tabu
            continue;
        assert(in<(::n));
        //2. perform move
        currClq[in] = 0;
        currClq[out] = 1;
        obj += bestContrib;
        if(obj>bestObj){
            bestObj = obj;
            if (obj>bstClqExisting)
              for(int i=0;i<(::n);i++)
                clique [i] = currClq[i];
        }
        tabu[in]  = iter+10+rand()%5;
        tabu[out] = tabu[in];
        tabu[out] = 0;
        //3. update contrib
        for(int i=0;i<(::n);i++){
            contrib[i] += ww[i][out];
            contrib[i] -= ww[i][in];
        }
    } 
    delete[] tabu;
    delete[] contrib;
    delete[] currClq;
    return bestObj;
}
//This function start trying cliques with maxCardinal, then maxCardinal - 1, ...  minCardinal
//Returns the best clique and records it in array clique
double maxWeightCliqueLocalSearch(int* clique, double * w, int maxCardinal, int minCardinal)
{
    double** ww = new double*[::n];  //::edge weights
    for(int i=0;i<(::n);i++){
        ww[i] = new double[::n];
        for(int j=0;j<(::n);j++)
            if(::edge[i][j])
                ww[i][j ]=(w[i]+w[j]); //divide it later by k-1, because a k-clique adds w[i] k-1 times in the k-1 ::edges linking i
            else
                ww[i][j] = -1000000;
        ww[i][i] = 0;
    }
    double cBest=0;
    double cLast=0;
    assert(maxCardinal<(::n));
    assert(maxCardinal>=minCardinal);
    assert(minCardinal>0);
    do{
        cLast = cBest;
        fastHeuristicBS(clique,w,maxCardinal);
        cBest = mainLSEngine(clique, ww, ITER_LOC_SEARCH,cLast*(maxCardinal-1));
        if(cLast==0)
            if(maxCardinal==minCardinal)
                if(cBest>=0) 
                    cLast = cBest/(maxCardinal-1);
        maxCardinal--;
        cBest = cBest / maxCardinal;
        if(cBest<0)
            cBest = 0;//failed
    }
    while((cBest>=cLast)&&(maxCardinal>minCardinal));
    maxCardinal ++;
    clog<<"heuristic clique size="<<maxCardinal<<endl;
    for(int i=0;i<(::n);i++)
        delete[] ww[i];
    delete[] ww;
    return cLast;
}
double maxWeightCliqueLocalSearch(int* clique, double * w, int maxCardinal)
{
    return maxWeightCliqueLocalSearch(clique, w, maxCardinal, 0);
}

/*-------------------------------------------------------------------------------------------------+
|                                                                                                  |
|                                         Graph readers                                            |
|                                                                                                  |
+-------------------------------------------------------------------------------------------------*/
int fileExists(char* filename){
    ifstream f(filename);
    return f.good();
}
void writeArray(int nn, double*v, char* filename, int precision)
{
    ofstream f (filename);
    if(!f.is_open()){
        cerr<<"Can not write to"<<filename<<"stop"<<endl;
        return;
    }
    for(int i=0;i<nn;i++)
        f<<setprecision(precision)<<v[i]<<" ";
    f.close();
}
void readArray(int nn, double*v, char* filename)
{
    ifstream f;
    f.open(filename);
    if ( f.fail() ){
           cerr << " Error opening " << filename << endl;
           return ;
    }
    for(int i=0;i<nn;i++)
        f>>v[i];
    f.close();
}
int readColFile(char* filename, int**& stables, int& colors){
     ifstream fic;
     fic.open(filename);
     if ( fic.fail() ){
            cerr << " Error opening " << filename << endl;
            return 0;
     }
     fic>>colors;
     if(colors<=0){
           cerr << "Found 0 colors on the first line of "<<filename<<endl;
           cerr << "I need to exit. This is not a proper coloring file\n";
           return 0;
     }
     int* stabSizes = new int[colors];
     stables   = new int*[colors];
     for(int i=0;i<colors;i++){
        stables[i] = new int[::n];
        for(int j=0;j<(::n);j++)
            stables[i][j] = 0;
        fic>>stabSizes[i];
     }
     for(int i=0;i<colors;i++)
        for(int j=0;j<stabSizes[i];j++){
            int vtx;
            fic>>vtx;
            stables[i][vtx] = 1;
        }
    delete[] stabSizes;
    fic.close();
    return 1;
}

void readGraph(char*filename){
     ifstream inFileStream;
     inFileStream.open(filename);
     if ( inFileStream.fail() ){
           cerr << " Error opening " << filename << endl;
           exit(0);
     }
     char token[100];
     inFileStream >> token;
     if (inFileStream.eof()){
           cerr << " Found EOF while reading in the beginning " << filename << endl;
           exit(0);
     }
     int x1, x2;     
     int mCount=0;
     while ( ! inFileStream.eof() ) {
        if (!strcmp(token, "p" )) {
            char dummyWord[50];
            inFileStream >> dummyWord >> ::n >> ::m;
            clog << "Vertices = " << ::n << ",";
            clog << "Edges = " << ::m << "\n";               
            ::edge=new int*[::n];
            for(int ii=0;ii<(::n);ii++){
                ::edge[ii]=new int[::n];
                for(int jj=0;jj<(::n);jj++)
                    ::edge[ii][jj]=0;
            }
        }                                  
        if (!strcmp(token, "e")){
            inFileStream >> x1 >> x2;
            x1--; x2--;
            if ( x1<0 || x2<0 || x1>=n || x2 >=n ){
                  cerr << "Error reading ::edges x1=" << x1 << ", x2=" << x2 << endl;
            }
            ::edge[x1][x2]=1;
            ::edge[x2][x1]=1;
            mCount++;
        }
        inFileStream >> token;
     }  
     if ( (::m != mCount))
           cerr<<"Warning reading graph. Announced m size=" << ::m << ", found=" << mCount  << endl;
     ::m = mCount;
     inFileStream.close();   
};
void readGraph(){
    ::n = 20;
    ::edge = new int*[::n];
    for(int i=0;i<n;i++){
        edge[i] = new int[::n];
        for(int j=0;j<(::n);j++)
            ::edge[i][j] = 0;
    }
    for(int i=0;i<n;i++)
        for(int j=0;j<i;j++)
            if((i-j)%4==0){
                ::edge[j][i]=1;
                ::edge[i][j]=1;
                ::m++;
            }
}
            
void testMaxStable()
{
    readGraph();
    printMat(::edge,::n);
    int*stab = new int[::n];
    double*w = new double[::n];
    for(int i=0;i<(::n);i++)
        w[i] = 1;
    cout<<maxWeightStableCplex(stab,w,1,2)<<endl;
    for(int i=0;i<(::n);i++)
        cout<<stab[i]<<" ";
    cout<<endl;
}
//Print a matrix on the output
void printMat(int**mat,int nn)
{
    for(int i=0;i<nn;i++){
        for(int j=0;j<nn;j++){
            clog.fill(' ');
            clog.width(2);
            clog<<mat[i][j];
        }
        clog<<endl;
    }
}
