/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include "graphstables.h"
#include "cliqueBBBS.h"
#include "dijkstra.h"
#include <cmath>
#include <cassert>
#include <ilcplex/ilocplex.h>                    

#define CUTS_HIJ_NOTLINKED_K 10         //max elems of an accepted cut in addCutEdg_hij_NotLinked_k
                                        //It is a constraint that can cut sol x_i=1/(largestClique)
#define LIM_SEARCH_HIJK 10000           //Stop if you found LIM_SEARCH_HIJK of hij not linked to k

#define SMOOTHED_RELAXED_CLIQUE_CUTPLAN //When solving by cutt-planes, use smoothed sols each iter
#define USE_HEURISTICS                  //Try first a heuristic in cut-planes, it can be my heuristic

#define EPSILON 1.0e-6
#define THREADS 0                       //The number of threads to use in cplex solver.  Let 0 the default and it 
                                        //will choose the nr of CPUs or of available hyper-threads of the machine
int fastClqSz = 0;
int maxClqElems = 0;
int calls = 0;                          //nr of inter calls

namespace{
//The model is shared among several calls to intersect, because the model
//already contains generated cuts associated to clique or odd holes
IloEnv         env;
IloModel       model(env);
IloCplex       cplex (env);
IloNumVarArray a;            //a=a/d^Ta,a[n]=\overline{\alpha} from Charnes Cooper
IloRangeArray  cuts(env);                     
int nbCuts = 0;
int*used;

//a[i]+a[j]+a[k]+a[l]+a[p]<=a[n]=freeTerm, ignore terms if j/k/l/p=-1
void addCliqueIneq(int i, int j, int k, int l, int p){
      IloExpr exprConstr(env);
      exprConstr=a[i]-a[n];
      used[i]++;
      if(j!=-1){
        exprConstr+=a[j];
        used[j]++;
      }
      if(k!=-1){
        exprConstr+=a[k];
        used[k]++;
      }
      if(l!=-1){
        used[l]++;
        exprConstr+=a[l];
      }
      if(p!=-1){
        used[p]++;
        exprConstr+=a[p];
      }
      cuts.add(exprConstr<=0);
      nbCuts++;
      model.add(cuts[nbCuts-1]);
      exprConstr.end();    
}
void addFastCliqueInequalities(int clqElems)
{
    double* oneMainVtx = new double[::n];
    int**reached       = new int*[::n];     //reached[i][j]=1 if j in clique of i
    for(int i =0;i<(::n);i++){
        reached[i] = new int[::n];
        for(int j=0;j<(::n);j++){
            reached[i][j] = 0;
            oneMainVtx[j] = n;
        }
        for(int j=0;j<i;j++)
            if(reached[j][i])
                oneMainVtx[j] = 1;
        oneMainVtx[i] = n*n;
        maxWeightCliqueBBBS(reached[i],oneMainVtx,0,clqElems);
        IloExpr exprConstr(env);
        exprConstr=-a[n];
        clog<<"fast clique for i="<<i<<":";
        for(int j=0;j<(::n);j++)
            if(reached[i][j]>0){
                exprConstr+= reached[i][j]*a[j];
                clog<<j<<" ";
            }
        clog<<endl;
        cuts.add(exprConstr<=0);
        nbCuts++;
        model.add(cuts[nbCuts-1]);
        exprConstr.end();    
    }
    for(int i=0;i<(::n);i++)
        delete[] reached[i];
    delete[] reached;
    delete[] oneMainVtx;
}
void addAllEdgeInequlities(){
    used = new int[::n];
    for(int i=0;i<(::n);i++)
        used[i] = 0;
    for(int i=0;i<n;i++)
        for(int j=i+1;j<n;j++)
            if(edge[i][j])
                addCliqueIneq(i,j,-1,-1,-1);
    for(int i=0;i<(::n);i++)
        if(used[i]==0){
            cerr<<"i not touched, adding x_i<=1 for i="<<i<<endl;
            addCliqueIneq(i,-1,-1,-1,-1);
        }
}
//do not make a vertex arise in more than enough cuts
void addCliqueInequalities(int maxLevel, int enough)
{
    assert(maxLevel<=5);       //I won't search cliques of size 6 or more
    assert(maxLevel>=2);       //please accept to always search ::edges (cliques size 2)
    used = new int[::n];
    for(int i=0;i<(::n);i++)
        used[i] = 0;

    for(int i=0;i<(::n);i++){
      if(used[i]>enough)
        continue;
      int edge_on_i_found = 0;
      for(int j=0;j<i;j++)
          if(::edge[i][j]){
              if(used[j]>enough)
                continue;
              edge_on_i_found = 1;
              int clique_3_found = 0;
              if(maxLevel>=3)
              for(int k=0;k<j;k++){
                  if(used[k]>enough)
                      continue;
                  if((::edge[i][k])&&(::edge[j][k])){
                      clique_3_found = 1;
                      int clique_4_found = 0;
                      if(maxLevel>=4)
                      for(int l=0;l<k;l++){
                          if(used[l]>enough)
                              continue;
                          if((::edge[i][l])&&(::edge[j][l])&&(::edge[k][l])){
                              clique_4_found = 1;
                              //cout<<"Clique of 4 found"<<i<<" "<<j<<" "<<k<<" "<<l<<" "<<"\n";
                              int clique_5_found = 0;
                              if(maxLevel>=5)
                              for(int p=0;p<l;p++){
                                 if(used[p]>enough)
                                      continue;
                                 if((::edge[i][p])&&(::edge[j][p])&&(::edge[k][p])&&(::edge[l][p])){
                                      clique_5_found = 1;
                                      addCliqueIneq(i,j,k,l,p);
                                      clog<<"Clique of 5 found"<<i<<" "<<j<<" "<<k<<" "<<l<<" "<<p<<"\n";
                                 }
                              }//end search of cliques of 5 with index p
                              if(!clique_5_found)
                                      addCliqueIneq(i,j,k,l,-1);
                          }
                      }//end search cliques of 4 with index l
                      if(!clique_4_found)
                            addCliqueIneq(i,j,k,-1,-1);
                  } 
              }//end search cliques of 3 with index k
              if(!clique_3_found){
                  if(maxLevel>=3)
                  for(int k=j+1;k<(::n);k++)
                      if(k!=i)
                          if((::edge[i][k])&&(::edge[j][k]))
                              clique_3_found = 1;
                  if(!clique_3_found)
                        addCliqueIneq(i,j,-1,-1,-1);
              }
          }//end if ::edge[i][j]
      if(!edge_on_i_found){
          for(int j=i+1;j<(::n);j++)
              if(::edge[i][j])
                  edge_on_i_found = true;
          if(!edge_on_i_found){
              //vertex not yet handled
              clog<<"i="<<i<<"disconnected!?"<<endl;
                addCliqueIneq(i,-1,-1,-1,-1);
          }
      }//end !edge_on_i_found
    }//end i
}
void addCliqueInequalities(int maxLevel)
{
    addCliqueInequalities(maxLevel,20);//do not make a vertex arise in more than 20 cuts
}
#ifdef CUTS_HIJ_NOTLINKED_K
int maxCutElems = CUTS_HIJ_NOTLINKED_K;
//Cut below are added only if they have at most maxCutElems non-zeros 
//Given edge ij and non edges ik and jk, set Ai=vertices connected to i
//We have xi<= xk+x(Ak-Ai) and xj<= xk+x(Ak-Aj) but also:
//xi+xj<=xj+x(Ak-intersect(Ai,Aj))=xj+x(z), i.e., z=Ak-intersect(Ai,Aj)
int addCutEdg_ij_NotLinked_k()
{
    int zMin = INT_MAX;
    int no_ijk=0;
    for(int i=1;i<(::n);i++)
      for(int j=i+1;j<(::n);j++)
        if(::edge[i][j])
          for(int k=0;k<(::n);k++)
            if( (!::edge[i][k]) && (!::edge[j][k]) && (i!=k) && (j!=k) ){
                no_ijk++;
                if(no_ijk>LIM_SEARCH_HIJK)
                    goto printfinal;
                assert(i!=j);
                int z = 0;
                for(int p=0;p<(::n);p++)
                  if( (::edge[k][p]) && ( !((::edge[i][p])&&(::edge[j][p]))) )
                      z++;
                if(z<=(::maxCutElems)){
                    IloExpr exprConstr(env);
                    exprConstr = a[i] + a[j] - a[k];
                    for(int p=0;p<(::n);p++)
                        if( (::edge[k][p]) && ( !((::edge[i][p])&&(::edge[j][p]))) ){
                            assert(k!=p);
                            exprConstr-= a[p];
                        }
                    zMin = min(zMin,z);
                    cuts.add(exprConstr<=0);
                    nbCuts++;
                    model.add(cuts[nbCuts-1]);
                    exprConstr.end();    
                }
            }
    printfinal:
    clog<<"nr_ijk="<<no_ijk<<endl;
    clog<<"zmin="<<zMin<<endl;
    return no_ijk;
}
int addCutEdg_hij_NotLinked_k()
{
    int zMin = INT_MAX;
    int no_hijk=0;
    for(int i=1;i<(::n);i++)
     for(int j=i+1;j<(::n);j++)
       if(::edge[i][j])
         for(int h=j+1;h<(::n);h++)
          if((::edge[i][h])&&(::edge[j][h]))
           for(int k=0;k<(::n);k++)
            if( (!::edge[i][k]) && (!::edge[j][k]) && (!::edge[h][k]) && (i!=k) && (j!=k) && (h!=k) ){
                no_hijk++;
                if(no_hijk>LIM_SEARCH_HIJK)
                    goto printfinal;
                assert(i!=j);
                int z = 0;
                for(int p=0;p<(::n);p++)
                  if( (::edge[k][p]) && ( !((::edge[i][p])&&(::edge[j][p])&&(::edge[h][p]))) )
                      z++;
                if(z<=(::maxCutElems)){
                    IloExpr exprConstr(env);
                    exprConstr = a[i] + a[j] + a[h] - a[k];
                    for(int p=0;p<(::n);p++)
                        if( (::edge[k][p]) && ( !((::edge[i][p])&&(::edge[j][p])&&(::edge[h][p]))) ){
                            assert(k!=p);
                            exprConstr-= a[p];
                        }
                    zMin = min(zMin,z);
                    cuts.add(exprConstr<=0);
                    nbCuts++;
                    model.add(cuts[nbCuts-1]);
                    exprConstr.end();    
                }
            }
    printfinal:
    clog<<"nr_hijk="<<no_hijk<<endl;
    clog<<"zmin gener="<<zMin<<endl;
    return no_hijk;
}
#endif


//Search an odd hole on which you can state \sum_{i\in I} b_i<=(|I|+1)/2
int finishRelaxStabByCutPlanesUsingOddHoles()
{
    IloNumArray currSol(env);
    int nrOddHoleCuts = 0;
    cplex.solve();
    cplex.getValues (a,currSol);
    double maxSolVal=currSol[0];
    for(int i=1;i<(::n);i++)
        if(currSol[i]>maxSolVal){
            maxSolVal = currSol[i];
            if(maxSolVal>0.4)
                break;
        }
    if(maxSolVal<=0.4)            //no Odd hole cut of size 5 can be violated when 
        return 0;                 //all values are <=0.4, because 5*0.4<=2 holds, same for 7*0.4<=3
    setGraphSize(2*::n);
    for(int i=0;i<(::n);i++)
        for(int j=i+1;j<(::n);j++)
            if(::edge[i][j]){
                //clog<<"i="<<i<<" j="<<j<<endl;
                //clog<<"currSol[n]FreeTerm="<<currSol[::n]<<endl;
                //clog<<"currSol[i]="<<currSol[i]<<endl;
                //clog<<"currSol[j]="<<currSol[j]<<endl;
                double weight = currSol[::n]-currSol[i]-currSol[j];
                if((weight<0)&&(weight>-EPSILON))
                    weight = 0;
                //clog<<weight<<endl;
                assert(weight>=0);
                addEdge(i,j+::n,weight);
                addEdge(j,i+::n,weight);
            }

    double*distances  = new double[2*(::n)];
    int* oddHole      = new int[2*(::n)];
    for(int i=0;i<(::n);i++){
        dijkstraCalcPaths(i,distances);
        //path from i to its copy i', this means cycle
        //|cycle|-2*sum_{i along cycle} xi <1 => violate sum_{i along //cycle}<=(|cycle|-1)/2
        if(distances[i+::n]<currSol[n]-EPSILON){                //odd hole found
           nrOddHoleCuts = 1;
           getVtxOnPathTo(i+::n,oddHole);
           double holeSize = 0;
           for(int jj=0;jj<(::n);jj++)
               if(oddHole[jj]+oddHole[::n+jj]>0)     //use jj or copy jj'
                    holeSize++;
           /* Avoid an "odd" hole like
                                +--+
                               >|  |
                              / +--+\
             +----+      +--+/       \
             |    | ---->|  |         \
             |  i |      |  |          |
             |    | <----|  |          |
             +----+      +--+<-\       v
                                \    +---+
                                 \   |   |
                                  \  |   |
                                   <-|   |
                                     +---+
           */
           if(((int)holeSize)%2==1){
                clog<<"ODD hole of len="<<holeSize<<"and value="<<distances[::n+i]<<endl;
                IloExpr exprConstr(env);
                exprConstr=-(holeSize-1.0)*0.5*a[::n];
                double currOddStatus = 0;
                for(int j=0;j<(::n);j++)
                    if(oddHole[j]+oddHole[::n+j]>0){
                        exprConstr+= a[j];
                        currOddStatus+=currSol[i];
                        clog<<"+hole["<<j<<"]";
                    }
                clog<<"<="<<(holeSize-1.0)*0.5*currSol[n];
                clog<<" (current="<<currOddStatus<<")"<<endl;
                 //exit(1);
                cuts.add(exprConstr<=0);
                nbCuts++;
                model.add(cuts[nbCuts-1]);
                exprConstr.end();    
            }
        }
    }
    delete[] distances;
    delete[] oddHole;
    dijkstraFreeMem();
    cplex.solve();
    return nrOddHoleCuts;
}

double solveMaxCliqueExact(int*newCut, double*sol, double cliqueMaxExisting){
    double cliqueMax=cliqueMaxExisting;
    //Yet again need correct numerical precision problems
    for(int i=0;i<(::n);i++)
        if((sol[i]<0)&&(sol[i]>-EPSILON))
            sol[i]=0;
    clog<<"... clique solver BB\n";
    //   double * test = new double[::n];
    //   for(int i=0;i<n;i++){
    //       test[i]=(int)floor(sol[i]*1000000);
    //       cout<<test[i]<<" ";
    //   }
    //   cout<<endl;
    //   clog<<cliqueMax<<endl;
    //   cliqueMax = maxWeightCliqueBBBS(newCut,sol,0,::maxClqElems);
    //cliqueMax = maxWeightCliqueCplex(newCut,sol,::maxClqElems);
    cliqueMax = maxWeightCliqueBBBS(newCut,sol,cliqueMax,::maxClqElems);
    //cliqueMax = maxWeightCliqueBBBS(newCut,sol,0,::maxClqElems);
    clog<<"    Finished clique solver\n";
    return cliqueMax;
}
double solveMaxClique(int* newCut, double*sol, int&finishedHeuristicPhase){
    #ifndef USE_HEURISTICS
        return solveMaxCliqueExact(newCut,sol,0);
    #endif
    double cliqueMax = 0;
    clog<<"    Launching clique solver in stab max Cut Planes";
    if(!finishedHeuristicPhase){
        cliqueMax = maxWeightCliqueLocalSearch(newCut,sol,::maxClqElems,::maxClqElems);

        clog<<"         Heuristic clique = "<<cliqueMax<<endl;
        assert(cliqueMax>=0);
          if(cliqueMax<=sol[n]+EPSILON){
            clog<<"         Finished with the heuristic clique finder for stab max. Going Exact";
            finishedHeuristicPhase = 1;
          }
    }
    if(finishedHeuristicPhase)                            //we tried setting lb=sol[n]
        return solveMaxCliqueExact(newCut,sol,cliqueMax); //sol[n] is \alpha in C-Cooper model
    return cliqueMax;
}
void finishRelaxStabByCutPlanesUsingClqCuts()
{
    IloNumArray currSol(env);
    double * queryPnt    = new double[::n+1];
    int * newCut    = new int[::n];
    double cliqueMax;
    int finishedHeuristicPhase = 0;
    int iterInternalCutPlanes = 0;
    #ifdef SMOOTHED_RELAXED_CLIQUE_CUTPLAN
    cplex.getValues (a,currSol);
    for(int i=0;i<=n;i++)
        queryPnt[i] = currSol[i];
    #endif
    do{
        iterInternalCutPlanes ++;
        if(iterInternalCutPlanes%15==0)
            finishedHeuristicPhase = 0;
        clog<<"start cplex solve with cuts="<<nbCuts<<endl;
        cplex.solve();
        clog<<"end cplex solve\n";
        cplex.getValues (a,currSol);
            
#ifndef SMOOTHED_RELAXED_CLIQUE_CUTPLAN
        for(int i=0;i<=n;i++)
            queryPnt[i] = currSol[i];
        cliqueMax = solveMaxClique(newCut,queryPnt,finishedHeuristicPhase);
#else
        for(int i=0;i<=n;i++)
            queryPnt[i] = 0.5*queryPnt[i] + 0.5*currSol[i];
        //solve clique max on smoothed currSol
        clog<<"Launching clique solver on smoothed query point"<<endl;
        solveMaxClique(newCut,queryPnt,finishedHeuristicPhase);
        //Recalc clique max because above returns the maxClique w. r. t. smoothed currSol
        cliqueMax= 0;
        for(int i=0;i<(::n);i++)
            if(newCut[i])
                cliqueMax+=currSol[i];
        if(cliqueMax<=currSol[n]+EPSILON){
            IloExpr exprConstr(env);
            exprConstr=-a[n];
            for(int i=0;i<n;i++)
                if(newCut[i]>0)
                    exprConstr+= newCut[i]*a[i];
            cuts.add(exprConstr<=0);
            nbCuts++;
            model.add(cuts[nbCuts-1]);
            exprConstr.end();    
            clog<<"Smoothed query point NOT enough. --------> Giving current queryPnt as query point for clique separator"<<endl;
            for(int i=0;i<=n;i++)
                queryPnt[i] = currSol[i];
            cliqueMax = solveMaxClique(newCut,queryPnt,finishedHeuristicPhase);
        }else{
            clog<<"Smoothed query point enough to sep curr queryPnt\n";
        }
#endif /*SMOOTHED_RELAXED_CLIQUE_CUTPLAN*/

        IloExpr exprConstr(env);
        exprConstr=-a[n];
        
        #ifndef NDEBUG
        double testCliqueMaxVal = 0;
        for(int i=0;i<(::n);i++)
            if(newCut[i])
                testCliqueMaxVal+=currSol[i];
        assert(abs(testCliqueMaxVal-cliqueMax)<EPSILON);
        #endif
        if(!finishedHeuristicPhase)
            clog<<"Heuristic phase ";
        clog<<" iterInternalCutPlanes="<<iterInternalCutPlanes<<" ";
        clog<<"cliquemax value="<<cliqueMax<<endl;
        for(int i=0;i<n;i++)
            if(newCut[i]>0){
                exprConstr+= newCut[i]*a[i];
                clog<<"+z["<<i<<"]*"<<newCut[i];
            }
        clog<<"<="<<currSol[n]<<endl;
        cuts.add(exprConstr<=0);
        nbCuts++;
        model.add(cuts[nbCuts-1]);
        exprConstr.end();    
    }while(cliqueMax>currSol[n]+EPSILON);
    cplex.solve();
    delete[] queryPnt;
    delete[] newCut;
}

//slower on some graphs, like r125.1c
double repeatedSeparation(double *x, double *direction, double*stabFrac){
    int * stab = new int[::n];
    int * stabNew = new int[::n];
    double * d = new double[::n];
    for(int i=0;i<(::n);i++)
        d[i] = x[i]+direction[i];
    double tStar = 1;
    double stabMax = maxWeightStableCplex(stab,d,1,20000); 
    while(stabMax>1+EPSILON){
        double xVal = 0; for(int i=0;i<n;i++)xVal+=x[i]*stab[i];
        double yVal = 0; for(int i=0;i<n;i++)yVal+=d[i]*stab[i];
        if(xVal>1+EPSILON){
            cerr<<"xval="<<xVal<<"Infeasible inner solution "<<endl;
            exit(1);
        }
        if(abs(stabMax-yVal)>EPSILON*10){
            cerr<<"Stab max="<<stabMax<<"and yVal="<<yVal<<"Should be equal "<<endl;
            exit(1);
        }
        tStar *= (1-xVal)/(yVal-xVal);
        cout<<"sep"<<tStar<<" "<<flush;
        for(int i=0;i<n;i++)
            d[i] = x[i]+(1-xVal)/(yVal-xVal)*(d[i]-x[i]);
        stabMax = maxWeightStableCplex(stabNew,d,1,10000); 
        if(stabMax>1+EPSILON)
            for(int i=0;i<n;i++)
                stab[i]=stabNew[i];
        //cout<<"stabMax="<<setprecision(12)<<stabMax<<endl;
    }
    cout<<"sep last done"<<endl;
    for(int i=0;i<n;i++)
        stabFrac[i] = stab[i];
    delete[] stab;
    delete[] stabNew;
    delete[] d;
    return tStar;
}

bool binarized = false; //did you already asked the model to use integer stables ?
double charnesCooperBB(double*& bst,double&,int);
//                1 - x^Ta
//return   min   ----------- because cuts are (x+ty)^T a <=1 for a=coloring
//                 d^T a
//x>=0, d arbitrary but non-zero 
#define ONE_EQ_CHARNES_COOPER 1
double intersect(double *x, double *d, double*fracStabl, int interStdStables)
{
    if(interStdStables>=1000) 
        return repeatedSeparation(x,d,fracStabl);
    #ifndef NDEBUG
    int onlyZeros = 1;
    for(int i=0;i<(::n);i++)
        if(d[i]!=0){
            onlyZeros = 0;
            break;
        }
    if(onlyZeros){
        cerr<<"You gave me a direction d only with zeros. Return INT_MAX but filling stable with zeros. \n";
        for(int i=0;i<(::n);i++)
            fracStabl[i]=0;
        return INT_MAX;
    }
    #endif
    calls ++;
    if(calls==1)
       a = IloNumVarArray(env, n+1, 0, 100000);  //a=a/d^Ta,a[n]=t from Charnes Cooper

    IloExpr expr(env);
    IloExpr expr2(env);
    for(int i=0;i<(::n);i++){
        expr+=-x[i]*a[i];
        expr2+=-x[i]*a[i];
    }
    expr+=a[n];
    expr2+=a[n];
    IloObjective obj(env, expr, IloObjective::Minimize);
    model.add(obj);
    if(calls==1)
        cout<<"Adding tStar>=0, this does not always speeds up things!\n";
    //model.add(expr2>=0);
    expr2.end();

    //d^T a = 1 from Charnes Cooper
    IloExpr exprConstr(env);
    for(int i=0;i<(::n);i++) 
        exprConstr += d[i]*a[i];
    IloRange equalCharnesCooper(exprConstr==ONE_EQ_CHARNES_COOPER);
    model.add(equalCharnesCooper);
    exprConstr.end();    
    if(calls==1){

      if(interStdStables){
          addAllEdgeInequlities();
          if(interStdStables%2==0){                       //int++ version
              cout<<"Attention: adding fast clique cuts\n";
              addFastCliqueInequalities(::fastClqSz);
          }
      }else{
          addCliqueInequalities(min(::maxClqElems,5));
          addFastCliqueInequalities(::fastClqSz);
          #ifdef CUTS_HIJ_NOTLINKED_K
           addCutEdg_ij_NotLinked_k();
           addCutEdg_hij_NotLinked_k();
          #endif
     }
    }

    #ifdef NO_CPLEX_OUTPUT
    cplex.setOut(env.getNullStream());
    cplex.setError(env.getNullStream());
    cplex.setWarning(env.getNullStream());
    #endif 
    cplex.extract(model);
    cplex.setParam(IloCplex::NumericalEmphasis, true);
    try{
        int nbHoleCuts =0 ;
        do{
            cplex.solve();
            if(interStdStables){  //projective cut planes with std stables
                if(!binarized){ //Useless to try to add cuts when all opt sols are std stables
                    finishRelaxStabByCutPlanesUsingClqCuts();
                    //Below might be useful when only solving one intersection
                    //nbHoleCuts = finishRelaxStabByCutPlanesUsingOddHoles();
                }
            }else{              //RR stables
                finishRelaxStabByCutPlanesUsingClqCuts();
                nbHoleCuts = finishRelaxStabByCutPlanesUsingOddHoles();
            }
        }while(nbHoleCuts>0);
    }catch (IloCplex::Exception e){
        cerr<<"Exception"<<e<<endl;
        if(cplex.getStatus()==IloAlgorithm::InfeasibleOrUnbounded){
            cerr<<"Attention: InfeasibleOrUnbounded. Strange\n";
        }
        cerr<<"Saving the model to trycatch.lp. Check it\n";
        cplex.exportModel("trycatch.lp");
    }

    double tStar      = cplex.getObjValue();
    //cout<<"tStar="<<tStar<<endl;exit(1);
    //retrieve the frac stable

    IloNumArray currSol(env);
    cplex.getValues (a,currSol);
    for(int i=0;i<(::n);i++){
        fracStabl[i] = currSol[i]/currSol[n];
    }
    //Verify
    #ifndef NDEBUG
        clog<<"tStarFresh="<<tStar<<endl;
        static double* hitPoint = new double[::n];
        double checkSum  = 0;
        double checkSumOnInfeas  = 0;
        double objFunc   = 0;
        double yTimesFrac= 0;
        double xTimesFrac= 0;
        double yTimesSol = 0;
        //clog<<"fracStable";
        //for(int i=0;i<(::n);i++)
        //    clog<<fracStabl[i]<<" ";
        //clog<<endl;
        for(int i=0;i<(::n);i++){
            yTimesFrac += fracStabl[i]*d[i];
            xTimesFrac += fracStabl[i]*x[i];
            hitPoint[i] = x[i]+tStar*d[i];
            checkSum   += hitPoint[i]*fracStabl[i];
            objFunc    += hitPoint[i];
            checkSumOnInfeas += (x[i]+d[i])*fracStabl[i];
            yTimesSol  += currSol[i]*d[i];
        }
        clog<<"x*FracStable="<<xTimesFrac<<endl;
        clog<<"yDir*FracStable="<<yTimesFrac<<endl;
        clog<<"freeTermVal="<<currSol[n]<<endl;
        clog<<"checkSumOnInfeas="<<checkSumOnInfeas<<endl;
        if(tStar<1-EPSILON)
            assert(checkSumOnInfeas>=1+EPSILON);
        assert(abs(yTimesSol-1)<EPSILON);
        assert(abs(checkSum-1)<EPSILON);
        clog<<"ObjValOnHitPoint="<<objFunc<<endl;
        if(calls==1)
            cplex.exportModel("test.lp");
    #endif /*NDEBUG*/
    if((interStdStables)&&(!binarized)){
        cout<<"tStar="<<tStar<<"---(going IP after "<<nbCuts<<" cuts)---"<<flush;
        if(interStdStables<10){//cplex version
            if(!binarized){
                //Add logical constraints
                for(int i=0;i<(::n);i++)                    
                    model.add( (a[i]<=0)+(a[i]>=a[n]) == 1);   //better for r125 and others
                    //model.add( (a[i]==0)+(a[i]==a[n]) == 1); //better for dsjc250.9 and le450_25.c quite alone
                binarized = true;
            }
            ////Do this if you want to insert them and then remove
            //IloConstraintArray zeroOrOne(env);
            //for(int i=0;i<(::n);i++)
            //    zeroOrOne.add( (a[i]<=0)+(a[i]>=a[n]) == 1);
            //model.add(zeroOrOne);
            //model.remove(zeroOrOne);
            cplex.solve();
            tStar = cplex.getObjValue();
            cplex.getValues (a,currSol);
            cout << "(cplex)";
        }
        //My B&B for the intersection sub-problem with standard stables
        if((interStdStables>=10)&&(interStdStables<20)){
            //Pay attention that below can return feasible sol that can be infeasible
            //as seen from the second call, this is normal given that the
            //branching can be different from call to call
            //This can be useful at least in mode "last projection with std stab"
            double* currSolTab=NULL;
            double tStarHeur = INT_MAX;
            int maxDepth = 4;
            charnesCooperBB(currSolTab,tStarHeur,maxDepth);
            if(tStarHeur>=INT_MAX-1){
                cerr<<"Exit: could not find clique with size="<<maxDepth<<endl;
                exit(1);
            }
            tStar = tStarHeur; 
            for(int i=0;i<(::n+1);i++)
                currSol[i] = currSolTab[i];
            delete[] currSolTab;
            cout << "(BBBS)";
        } 

        cout << "----(IP)---->"<< tStar <<" stab start=";
        int printed=0;
        for(int i=0;(i<(::n))&&(printed<n);i++){
            fracStabl[i] = currSol[i]/currSol[n];
            if(fracStabl[i]>0.1){
                printed++;
                cout<<i<<"/";
                cout<<currSol[i]<<" ";
            }
        }
        cout<<":"<<currSol[n];
        cout<<endl;
    }
    //Remove only the constraints that change for next intersection
    model.remove(obj);
    model.remove(equalCharnesCooper);

    return tStar;
}
double distFromBounds(double val, double uval){
    if(val<EPSILON)
        return 0;
    if(val>uval-EPSILON)
        return 0;
    double v1 = val/uval;
    return max(v1,1-v1);
}
double bstFeasible;
double charnesCooperBB(double*& bst, double&tStarHeur, int maxDepth){
    static int depth=0; depth++;
    if(depth>maxDepth){
        depth--;
        return INT_MAX;
    }
    if(bst==NULL){
        bst = new double[::n+1];
        bstFeasible = INT_MAX;
        for(int i=0;i<=n;i++)
            bst[i] = 0;
    }
    cplex.solve();
    //finishRelaxStabByCutPlanesUsingClqCuts(); cplex.solve();
    IloNumArray currSol(env);
    cplex.getValues (a,currSol);
    double currVal = cplex.getObjValue();
    if(depth==maxDepth)
        if(currVal<tStarHeur){
            tStarHeur = currVal;
            for(int i=0;i<=n;i++)   
                bst[i] = currSol[i];
        }
    if(currVal>=bstFeasible){
        depth--;
        return INT_MAX;
    }
    int    mostFractional      = -1;
    double mostFractionalValue = 0;
    for(int i=0;i<n;i++){
        double dstToBounds = distFromBounds(currSol[i],currSol[n]);
        if(dstToBounds>mostFractionalValue){
            mostFractional = i;
            mostFractionalValue = dstToBounds;
        } 
    }
    if(mostFractionalValue==0){
        if(currVal<bstFeasible){
            bstFeasible = currVal;
            if(bstFeasible<tStarHeur){
                tStarHeur = bstFeasible;
                for(int i=0;i<=n;i++)   
                    bst[i] = currSol[i];
            }
        }
        cout<<"*"<<currVal<<"*";
        depth--;
        return currVal;
    }
    //cout<<"branching on "<<mostFractional<<" with "<<currSol[mostFractional]
    //                   <<"/"<<currSol[n]<<":"<<mostFractionalValue;
    cout<<depth<<" "<<flush;

    double branch1, branch2;

    //Do not need testing you can assing mostFractional! You never have a fractional value on mostFractional
    //if there is an edge between mostFractional and an already selected vtx
    IloExpr exprConstrSec(env);
    exprConstrSec = a[mostFractional] - a[n];
    IloRange secBranch(exprConstrSec==0);
    model.add(secBranch);
    branch2 = charnesCooperBB(bst,tStarHeur,maxDepth);
    model.remove(secBranch);
    exprConstrSec.end();    

    IloExpr exprConstrFirst(env);
    exprConstrFirst = a[mostFractional];
    IloRange firstBranch(exprConstrFirst==0);
    model.add(firstBranch);
    branch1 = charnesCooperBB(bst,tStarHeur,maxDepth);
    model.remove(firstBranch);
    exprConstrFirst.end();    

    depth --;
    return min(branch1,branch2);
}
}//end unnamed namespace

//these functions should be visible from outside
double intersectRRStables(double *x, double *d, double*fracStabl){
    return intersect(x,d,fracStabl,0);
}
double intersectStdStables(double *x, double *d, double*fracStabl, int verCode){
    return intersect(x,d,fracStabl,verCode);
}
