/*---------------+---------------------------------------------------------------+--------------+
                 | Author: Daniel Porumbel 2018       daniel.porumbel@cnam.fr    |
                 |License: Creative Commons Attribution 3.0 Unported License     |
                 |         http://creativecommons.org/licenses/by/3.0/           |             
                 +--------------------------------------------------------------*/
#include "general.h"
#include "graphstables.h"
#include "intersect.h"
#include "CuttingPlanesEngine.h"
#include "cliqueBBBS.h"
#include <iostream>
#include <climits>
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <cstring>
#include <cassert>
using namespace std;
/*======================   Project-wide Global Variables     =====================*/
int**   edge;        //Almost all functions work on the same graph
int     n;
int     m;
#define EPSILON 1.0e-6  

/*======================   Global Variables Used in Main     =====================*/

#define TIMEOUT 15000
#define LST_IT_PROJ_STD_STB -7
#define INTER_GENERAL_VER    1
#define INTER_WITH_BBBS_BB   11
#define INTER_PLUS_PLUS      2        //decrease alpha when inter sub returns t*=0
                                      //+ use fast clique inequalities
int inter_with_std_stables =0;        //if >0, do not use RR-stables for proj
                                      // cut planes, but use std stables for 
                                      //inter_with_std_stables iters
bool cplexPricingSep = false;         //by defult BBBS pricing

int    iters        = 0;
double bestLowerBnd = 0;
double lBndCurrIter = 0;
double firstFarley  = 0;
double startTime;
int    kInitColoring;        //the number of colors of the heuristic solution from file
double * bestFeasib = NULL;
double alphaInit9999    = 0.9999;//xnew=x+0.9999*t*d
int    useStdStabMaxSz=0;    //the max number of vtx in a stable, useful for BBBs in std method
                             //0 means by default we use RR-stable stables in projCutPlanes
                             //method (separatorInter)
void correctPrecisionProbls(double *x)
{
    //correct for precision problems for the duals x that can have x[i]=-EPSILON
    for(int i=0;i<n;i++){
        if(x[i]<-EPSILON){
            cerr<<"I need to exit because of a really negative dual for i="<<i<<":"<<x[i]<<endl;
            exit(1);
        }
        if(x[i]>1+EPSILON){
            cerr<<"I need to exit because of a dual>1+EPSILON for i="<<i<<":"<<x[i]<<endl;
            exit(1);
        }
        if(x[i]<EPSILON)    //This rounding simplifies any max stable solve 
            x[i]=0;
        if(x[i]>1-EPSILON)  //This rounding simplifies any max stable solve 
            x[i]=1;
    }
}


//return violation = rightHandVal - newCut^\top x
double separatorInter (const int nrVars, double*x, double * newCut, double&rightHandVal)
{ 
    assert(nrVars==::n);                           //number of vertices
    correctPrecisionProbls(x);

    if(bestFeasib==NULL){
        bestFeasib = new double[::n];
        for(int i=0;i<(::n);i++)
            bestFeasib[i] = 0.000;
    }
    double * direction = new double[::n];
    for(int i=0;i<nrVars;i++)
        direction[i] = x[i]-bestFeasib[i];

    /*============      LAUNCHING INTERSECTION AND CALC DIRECTIONS      ==============*/
    clog<<"**********Launching intersection***************\n";
    double tStar;
    if(inter_with_std_stables>0)
        tStar = intersectStdStables(bestFeasib,direction,newCut,inter_with_std_stables);
    else
        tStar = intersectRRStables(bestFeasib,direction,newCut);
    clog<<"tStar="<<tStar<<endl;
    if(tStar<EPSILON)
        if(inter_with_std_stables == INTER_PLUS_PLUS)    
            alphaInit9999*=0.99;

    lBndCurrIter = 0;
    for(int i=0;i<(::n);i++)
        lBndCurrIter+=bestFeasib[i]+tStar*direction[i];
    if(lBndCurrIter>bestLowerBnd)
        bestLowerBnd = lBndCurrIter;
    
    rightHandVal     = 1;
    double violation = 1;
    for(int i=0;i<(::n);i++)
        violation -= newCut[i] * x[i];

    for(int i=0;i<(::n);i++){
        bestFeasib[i] = bestFeasib[i] + tStar*direction[i];
        bestFeasib[i]*=alphaInit9999;
    }

    //Code to print solutions and investigate the bang bang effects.
    //static int printed=0;
    //if(printed%10==0){
    //    for(int i=0;i<n;i++)
    //        cout<<setw(5)<<left<<setprecision((bestFeasib[i]>1000)?4:(bestFeasib[i]>0.1?3:2))<<bestFeasib[i]<<" ";
    //    cout<<"TT"<<endl;
    //}
    //printed++; if(printed==100) exit(1);
    
    /*============                PRINT INFO AND RETURN                 ==============*/
    cout<<"Iter "<<setw(3)<<(iters+1)<<" tStar="<<setw(9)<<setprecision(3)<<tStar
                 <<" ubnd="<<setw(9)<<setprecision(7)<<::upperBound<<"  lbnd="
                 <<setw(8)<<lBndCurrIter<<" BEST_LB="<<setw(8)<<bestLowerBnd
                 <<"         Tm="<<getCPUTime()-startTime<<endl;

//    if(::upperBound-floor(::bestLowerBnd)-1<EPSILON)
//        return INT_MAX;                               //signifies gap closed, stop
    #ifdef TIMEOUT
        if(getCPUTime()-startTime >= TIMEOUT){
            cerr<<"ATTENTION: reached TIMEOUT="<<TIMEOUT<<" sec. Signaling gap closed to stop it all, but the gap is not really closed!!!\n";
            return INT_MAX;
        }
    #endif
    delete[] direction;
    iters++;
        
    #ifndef NDEBUG
    clog<<"xViolation="<<violation<<endl;
    clog<<"Adding:";
    for(int i=0;i<nrVars;i++){
        if(newCut[i]>0)
            clog<<newCut[i]<<"*x["<<i<<"]+";
    }
    clog<<"<="<<rightHandVal<<" with violation="<<violation<<"\n";
    #endif
    return violation;
}


//return violation = rightHandVal - newCut^\top x
double separatorStable (const int nrVars, double*x, double * newCut, double&rightHandVal)
{ 
    assert(nrVars==::n);                           //number of vertices
    correctPrecisionProbls(x);

    rightHandVal = 1;    
    int*stable=new int[::n];
    double stabMax;
    clog<<"Launched stable max solver \n";


    /*========                     SOLVING STABLE MAX                        =========*/
    if((bestLowerBnd==0)&&(::m<50000)){ //the very first iteration
        if(!cplexPricingSep)
            stabMax = maxWeightStableCplex(stable,x,1,25);//1st try the Cplex LP solver 25 secs
        else
            stabMax = maxWeightStableCplex(stable,x,1,0); //last 0=no timeout
        cout<<"First stabMax="<<stabMax<<endl;
        if(stabMax<=0){                 //If cplex timeouts and returns a heuristic
            if(stabMax>=-1){            //if it does not have time to cut duals x
                clog<<"We are the very first iteration and I need/run cliquer Ostergard\n";
                stabMax= maxWeightStableBBBS(stable,x,0,::useStdStabMaxSz);
            }else{                      //if cplex found a heuristic that cuts x
                //solve the LP relaxation => stabMax will be larger than a binary stabMax
                //but the relaxation does not write a stable,  we take the non-optimal stable, 
                stabMax = maxWeightStableCplex(NULL,x,0,50); 
                cout<<"stabMax "<<stabMax<<" computed by linear relaxation, but take best heuristic"
                                         <<" stable to generate a column\n";
            }
        }
    }
    else{                                           //normal iters use my BBBS solver
        double tStartCliquer = getCPUTime();
        if(cplexPricingSep)
            stabMax = maxWeightStableCplex(stable,x,1,0); //last 0=no time out
        else
            stabMax= maxWeightStableBBBS(stable,x,0,::useStdStabMaxSz);

        if(getCPUTime()>tStartCliquer+2000){        //stop col gen if stabMax needs >1000 seconds
            cerr<<"The problem becomes too difficult for me. Need >2000 secs for stab max. I'll stop\n";
            return INT_MAX;
        }
    }


    /*=================          CALCULATING VIOLATION             ===================*/
    double violation = rightHandVal;
    clog<<"Adding:";
    for(int i=0;i<nrVars;i++){
        newCut[i] = stable[i];
        violation -= stable[i]*x[i];
        if(stable[i])
            clog<<"x["<<i<<"]+";
    }
    clog<<"<=1 with violation="<<violation<<"\n";

    /*===================        UPDATING BOUNDS, PRINTING     =======================*/
    double lBndCurrIter = upperBound/stabMax;
    if(bestLowerBnd==0)
        firstFarley= lBndCurrIter;
    if(lBndCurrIter>bestLowerBnd)
        bestLowerBnd = lBndCurrIter;
    clog<<"CG obj:"<<::upperBound<<" lb here="<<lBndCurrIter
          <<" BEST="<<bestLowerBnd
          <<"          ITER="<<iters<<endl;
    iters++;

    cout<<"Iter "<<setw(3)<<iters
        <<" ubnd="<<setw(9)<<setprecision(7)<<::upperBound<<"  lbnd="
        <<setw(8)<<lBndCurrIter<<" BEST_LB="<<setw(8)<<bestLowerBnd
        <<"         Tm="<<getCPUTime()-startTime<<endl;

    if(::upperBound-floor(::bestLowerBnd)-1<EPSILON)
        return INT_MAX;                               //signifies gap closed, stop

    #ifdef TIMEOUT
        if(getCPUTime()-startTime >= TIMEOUT){
            cerr<<"ATTENTION: reached TIMEOUT. Signaling gap closed to stop it all, but the gap is not really closed!!!\n";
            return INT_MAX;
        }
    #endif
    for(int i=0;i<(::n);i++)
        clog<<x[i]<<" ";
    clog<<endl;
    #ifndef NDEBUG
    for(int i=0;i<(::n);i++)
       for(int j=0;j<(::n);j++)
            if(edge[i][j]){
                if((stable[i])&&(stable[j])){
                    cerr<<"Stable solver not good. It put "<<i<<" and "<<j<<"in \
                           the stable, but they are connected\n";
                }
            } 
    #endif
    delete[] stable;
    return violation;
}


double separator (const int nrVars, double*x, double * newCut, double&rightHandVal)
{ 
    if(::useStdStabMaxSz)
        return separatorStable(nrVars,x,newCut,rightHandVal);
    else
        return separatorInter(nrVars,x,newCut,rightHandVal);
}


//Returns an integer parameter from the config file, e.g., the value of kMax
//Default return value: INT_MAX (when requestedParam is not found)
int getParamFromConfigFile(char* requestedParam)
{
    if(rindex(requestedParam,'/')!=NULL)
        requestedParam = rindex(requestedParam,'/')+1;
    ifstream fConfig;
    char oneWord[100];
    char configFile[100] = "init.cfg";
    fConfig.open(configFile);
    string s;
	if(fConfig.good()){
		while(!fConfig.eof()){
			fConfig >> oneWord;
			if(!strcmp(requestedParam,oneWord)){
				fConfig >> oneWord;
				int toRet;
				sscanf(oneWord,"%d",&toRet);
				fConfig.close();
                cout<<requestedParam<<" set to "<<toRet<<" using the value from "<<configFile<<endl;
				return toRet;
			}		
		}
	fConfig.close();		
	}	
	cerr<<"You can insert in (in "<<configFile<<") a line with the form: '"<<requestedParam<<" <VALUE>' in case you want to specify a value for this parameter."<<endl;
	return INT_MAX;
    
}

//Linux kernel coding rules: macros resembling
//functions may be named in lower case
#define printClqSzForWhichMaxCliqueNeeds10MilliSecs(graphfile)                           \
  do{                                                                                    \
    readGraph(graphfile);                                                                \
    ::fastClqSz = getParamFromConfigFile(graphfile);                                     \
    if(::maxClqElems*10>::n*9){                                                          \
        ::maxClqElems = (int)(0.9*(double)n);                                            \
        cerr<<"maxClqElems>=0.9n, strange. I reduce it to 0.9n="<<maxClqElems<<endl;     \
    }                                                                                    \
    int maxDegree = 0;                                                                   \
    int*stab = new int[::n];                                                             \
    double* ones = new double[::n];                                                      \
    double* zeros = new double[::n];                                                     \
    double nrEdg = 0;                                                                    \
    for(int i=0;i<n;i++) {                                                               \
        ones[i]    = 1;                                                                  \
        zeros[i]   = 0;                                                                  \
        int degree = 0;                                                                  \
        for(int j=0;j<n;j++)                                                             \
            degree+=::edge[i][j];                                                        \
        maxDegree= max(maxDegree,degree);                                                \
        nrEdg   += degree;                                                               \
    }                                                                                    \
    assert((int)nrEdg%2==0);                                                             \
    assert(nrEdg/2==::m);                                                                \
    cout<<"Density"<<nrEdg/(::n*(::n-1))<<endl;/*m/(n(n-1)/2)*/                          \
    cout<<"maxDegree="<<maxDegree<<endl;                                                 \
    /*cout<<"maxCliqueLocSearch"<<maxWeightCliqueLocalSearch(stab,ones,::n/2)<<endl;*/   \
    int k;                                                                               \
    for(k=3;k<(::n);k++){                                                                \
        double startCliqueBB = getCPUTime();                                             \
        int kTest = maxWeightCliqueBBBS(stab,ones,0,k);                                  \
        double endCliqueBB   = getCPUTime();                                             \
        cout<<"Time Clique BB for k="<<k<<": "<<(endCliqueBB-startCliqueBB)<<endl;       \
        if(k!=kTest)                                                                     \
            cout<<"No clique of size "<<k<<endl;                                         \
        if((k!=kTest)||endCliqueBB-startCliqueBB>0.01){                                  \
            k--;                                                                         \
            break;                                                                       \
        }                                                                                \
    }                                                                                    \
    cout<<"Speedy k="<<k<<endl;                                                          \
    if(::fastClqSz==INT_MAX)                                                           \
        ::fastClqSz = 5;                                                               \
    if(::maxClqElems==0)                                                                 \
        ::maxClqElems = ::fastClqSz;                                                   \
    cout<<"Computing relaxed stable with maxClqElemsForLPSTable="<<::maxClqElems<<"... ";\
    cout<<flush;                                                                         \
    double*dummy = new double[n];                                                        \
    double tStar = intersectRRStables(zeros,ones,dummy);/* runs BBBS with weights=ones*/ \
    cout<<"Done: =>maxRelaxedStable="<<1/tStar<<"     for tStar="<<tStar<<endl;          \
    /*cout<<"maxStableCplex="<<maxWeightStableCplex(stab,ones,1,20)<<endl;*/             \
    /*cout<<"maxStableBBBS="<<maxWeightStableBBBS(stab,ones,34,::n)<<endl;*/             \
  }while(false);


void oneAdditionalIter(char* graphName){
    char* fileForLastInner=new char[1000];
    sprintf(fileForLastInner,"%s.%d.last",graphName,::maxClqElems);
    if(bestFeasib==NULL){
        bestFeasib = new double[n];
        if(fileExists(fileForLastInner)){
            readArray(::n, bestFeasib,fileForLastInner);
            //for(int i=0;i<n;i++) if(bestFeasib[i]<0.00000001) bestFeasib[i]=0;
            cout<<"Loaded inner solution from "<<fileForLastInner<<endl;
        }else{
            cout<<"I understand you want one projection but I could not"
                  "find input inner solution in "<<fileForLastInner<<endl;
            cout<<"Think of using option -k to take inner sol from file\n";
            if(::maxClqElems==0)
                cout<<"The 0 in the filename comes from not providing -k CLQSZ\n";
            cout<<"I'l project from 0 0 0...0\n";
            for(int i=0;i<n;i++)
                bestFeasib[i] = 0;
        }
    }else{
        if(fileExists(fileForLastInner))
            cout<<"I wont save last inner sol because the following file "
                  "already exists "<<fileForLastInner<<endl;
        else{
            cout<<"I'll write last inner sol to "<<fileForLastInner<<endl;
            writeArray(::n, bestFeasib,fileForLastInner,9);
        }
        cout<<"An additional projection considering std (non-RR) stables"<<endl;
    }
    inter_with_std_stables = INTER_GENERAL_VER;
    double * lastDirOnes = new double[::n];
    cout<<"Base of projection:";
    //if(maxClqElems>=10000)  maxClqElems = 5;//dont use cliquer for inter-stab
    double valProj = 0;
    for(int i=0;i<n;i++){
        lastDirOnes[i] = 1;
        //bestFeasib[i] /= 0.9999;//Do not renormalize, better not touch facets
        //stringstream tmp; tmp << setprecision(6) << fixed << bestFeasib[i]; 
        //sscanf(tmp.str().c_str(),"%lf",&bestFeasib[i]);
        if(bestFeasib[i]>0.01)
            cout<<i<<"/"<<setprecision(9)<<bestFeasib[i]<<" ";
        //if(bestFeasib[i]>0.5){ for(int j=0;j<i;j++)if(bestFeasib[j]>0.5)cout<<"edge to previous selected="<<edge[i][j]<<"\n"; }
        valProj += bestFeasib[i];
    }
    cout<<endl<<endl;
    cout<<"Val of projected point="<<valProj<<endl;
    double * lastCut = new double[::n];
    double rHandLast ;
    calls=0;
    separatorInter (n, lastDirOnes, lastCut, rHandLast);
    //for(int i=0;i<n;i++) if(bestFeasib[i]>0.5)bestFeasib[i]=200;else bestFeasib[i]=1;
    cout<<endl;
    cout<<"Checking feasibility of the following last pierce point:";
    for(int i=0;i<n;i++){
        //do not need, because no multipl with 0.9999 in this mode: bestFeasib[i] /= 0.9999;
        if(bestFeasib[i]>0)
            cout<<i<<"/"<<bestFeasib[i]<<" ";
    }
    cout<<"\nI search for stab max with weights given by last solution..."<<flush;
    int* lastStab= new int[n];
    double smax;

    char * fname_plus_stabmax = new char [300];
    sprintf(fname_plus_stabmax,"%sstabmax",graphName);
    ::useStdStabMaxSz = getParamFromConfigFile(fname_plus_stabmax);
    if(::useStdStabMaxSz<15){
        cout<<"use BBBS"<<flush;
        smax = maxWeightStableBBBS(lastStab,bestFeasib,::useStdStabMaxSz);
    }
    else{
        //cout<<"Use cliquer"<<flush;
        //smax = maxWeightStable_UseCliquerOstergard(lastStab,bestFeasib);
        //cout<<"stab max cliquer="<<smax<<endl;exit(1);
        cout<<"use cplex"<<flush;
        smax = maxWeightStableCplex(lastStab,bestFeasib,1,222);
    }
    cout<<"\nFound Stab max="<<smax<<flush;
    cout<<endl<<"I recount stab max from reported stable=";
    smax=0;
    for(int i=0;i<n;i++){
        if(!((lastStab[i]==0)||(lastStab[i]==1)))
            cerr<<"Last stable contains a non binary digit at index:"<<i<<endl;
        if(lastStab[i]){
            cout<<i<<"";
            smax+=bestFeasib[i];
            cout<<"("<<setprecision(9)<<smax<<") ";
        }
    }
    cout<<endl;
    cout<<"stabmax="<<smax<<endl;
}
void getMethod(char* code, char* graphName){
    if( !strncmp(code,"one",3)){
        cout<<"I will perform only one projection towards 1 1 1.."<<endl;
        cout<<"***********************************************************"<<endl;
        oneAdditionalIter(graphName);
        exit(EXIT_SUCCESS);
    }
    if( !strncmp(code,"stdc",4)){
        ::cplexPricingSep    = true;
        ::useStdStabMaxSz = 1000;
        cout<<"Std CG with standard stables. Pricing solved via cplex only"<<endl;
        return;
    }
    if( !strncmp(code,"std",3)){
        ::useStdStabMaxSz = 1000;
        cout<<"Std CG with standard stables. Pricing solved via BBBS (or cplex at iter 1)"<<endl;
        return;
    }
    if( !strncmp(code,"intrr",5)){
        cout<<"Inter/proj cut planes with RR stables";
        if(code[5]=='l'){
            inter_with_std_stables = LST_IT_PROJ_STD_STB;
            cout<<"I will do a last proj with std stab"<<endl;
        }
        return;
    }
    if( !strncmp(code,"int",3)){
        inter_with_std_stables = INTER_GENERAL_VER ;
        if(code[3]=='s')
            inter_with_std_stables = 1000 ;
        if(code[3]=='+'){
            cout<<"You chose improved proj cut planes\n";
            inter_with_std_stables = INTER_PLUS_PLUS;
        }
        cout<<"Inter/Proj cut planes with std stables\n";
        if(alphaInit9999==0.9999){
            cout<<"You did not set alpha for xnew=x+alpha*tStar*d. It is 1\n";
            alphaInit9999 = 1;
        }else{
            cout<<"Atention: I won't use alpha=1, but alpha="<<alphaInit9999<<endl;
        }
        return;
    }
    if(!strncmp(code,"cl",2)){
        printClqSzForWhichMaxCliqueNeeds10MilliSecs(graphName);
        exit(EXIT_SUCCESS);
    }
    cerr<<"I could not understand method code (first arg) "<<code<<endl;
    cerr<<"Type ./main to see acceptable methods\n";
    exit(EXIT_FAILURE);
}
int main(int argc, char**argv){
    #ifdef NDEBUG
    clog.setstate(ios_base::failbit);       //disable clog
    #endif
    char* colFile=NULL;
    int warmStartFromCol = 1;
    if(argc<3){
        cerr<<"Main usage: ./main METHOD graphDimacs.col [startColoring.clr] [-si[lent]] "
              "[-kMaxCliqueCardToSolveLPProjModelByBBBS] [-alphaDIGITS, ex -alpha55 for 0.55] "
              "[-no[StartFromColor]] \n\n"
              "METHOD can be:\n"
              "    int       : projective cut planes with standard stables\n"
              "    int++     : same as above, but improved with fast clique cuts and flexible alpha, i.e., decrease it when t*=0\n"
              "    intrr     : projective cut planes with RR stables (faster sub-problems, lower quality bounds)\n"
              "    intrrlast : proj cut planes with RR stab+one last iter \n"
              "    std       : standard column gener with standard stables and BBBS pricing\n" 
              "    stdc[plex]: standard column gener with standard stables and cplex pricing\n" 
                               //"with standard stables projecting to 1 1 ... 1 1\n"
              "    one       : one projection towards 1 1 ... 1 only \n"
              "    cl        : finds max size for which BBBS needs <0.01 secs\n"
              "\n Attention: It tries to take a start legal coloring from graphDimacs.col.clr\n";
        return EXIT_FAILURE;
    }
    while(argv[argc-1][0]=='-'){
        int argcAll = argc;
        if( (argv[argc-1][0]=='-')&&(!strncmp(argv[argc-1],"-no",3))){
            warmStartFromCol = 0;
            cout<<"I won't start from an existing coloring"<<endl;
            argc--;
        }
        if( (argv[argc-1][0]=='-')&&(!strncmp(argv[argc-1],"-alpha",6))){
            int digits = atoi(argv[argc-1]+6);
            if(digits==0){
                alphaInit9999 = 0;
                cout<<"Using intersection method 'shooting' from zero "
                      "(old intersection version)"<<endl;
            }else{
                int noDigits    = (int)log10((double)digits)+1;
                double powerTen = pow(10, noDigits);
                alphaInit9999 = ((double)digits)/((double)powerTen);
                cout<<"I'll define xnew=alpha*x+alpha*tStar*d, with alpha="
                    <<alphaInit9999<<endl;
            }
            argc--;
        }
        if( (argv[argc-1][0]=='-')&&(argv[argc-1][1]=='k')){
            ::maxClqElems = atoi(argv[argc-1]+2);
            cout<<"Cmd line maxClqElemsForLPSTable="<<::maxClqElems<<endl;
            if(::maxClqElems<2){
                cerr<<"You asked cliques of size<2, impossible!";
                exit(EXIT_FAILURE);
            }
            argc--;
        }
        if( (argv[argc-1][0]=='-')&&(!strncmp(argv[argc-1],"-si",3))){
            cout<<"I'll be silent and disable clog!"<<endl;
            clog.setstate(ios_base::failbit);       //disable clog
            argc--;
        }
        if(argcAll==argc){
            cerr<<"I could not understand the '-' option '"<<argv[argc-1]<<"'.\n";
            return(EXIT_FAILURE);
        }
    }
    //Left with max 4 args including main,eg, ./main int graphDimacs.col [startCol.clr]
    if(argc>4){
        cerr<<"Could not parse command line. More than 3 arguments without '-'. Type ./main for help\n";
        return EXIT_FAILURE;
    }
    if(argc<4){
        colFile = new char[500];
        sprintf(colFile,"%s.clr",argv[2]);
    } else
        colFile = argv[3];

    readGraph(argv[2]);
    cout<<"***********************************************************"<<endl;
    cout<<"Running the method below:\n";
    getMethod(argv[1],argv[2]);
    cout<<"***********************************************************"<<endl;

    if((::maxClqElems<1000)&&(::maxClqElems*10>::n*9)){
        ::maxClqElems = (int)(0.9*(double)n);
        cerr<<"\n WARNING: maxClqElems>=0.9n, strange. I reduce it to 0.9n="<<maxClqElems<<endl;
    }

    if(::useStdStabMaxSz){
        char * fname_plus_stabmax = new char [300];
        sprintf(fname_plus_stabmax,"%sstabmax",argv[2]);
        ::useStdStabMaxSz = getParamFromConfigFile(fname_plus_stabmax);
        if(::useStdStabMaxSz==INT_MAX)
            ::useStdStabMaxSz=::n;
        cout<<"Consider the max stable cardinal is"<<::useStdStabMaxSz<<endl;
        delete[] fname_plus_stabmax;
    }else{
        ::fastClqSz = getParamFromConfigFile(argv[2]);
        if(::fastClqSz==INT_MAX)
            ::fastClqSz = 5;
        cout<<"fastClqSz (highest size for which BBBS (solving max weighted clique) needs <0.01s)="<<::fastClqSz<<endl;
        if(::maxClqElems==0){
            if(inter_with_std_stables){
                ::maxClqElems = 2;
                cout<<"You did not indicate k, I take 2, i.e., I won't use cuts\n";
            }else{
                ::maxClqElems = ::fastClqSz;
                cout<<"You did not indicate k, I take the below (fastClqSz) value for"
                      "which BBBS needs <0.01 secs\n";
            }
        }
        cout<<"maxClqElemsFor (k) to define reinforced LP Relaxed Stable (RR-stables)="<<::maxClqElems<<endl;
    }
    

    double *b;            //objective function coefficients
    b = new double[n];
    for (int i=0;i<n;i++)
        b[i] = 1; 

    CuttingPlanesEngine cutPlanes(n,separator);      
    cutPlanes.setVarBounds(0,1);
    cutPlanes.setObjCoefsMaximize(b);

    int**stablesInitColoring = NULL;
    if((warmStartFromCol) && (readColFile(colFile,stablesInitColoring,kInitColoring))){
        cout<<"I took a legal coloring with k="<<kInitColoring<<endl;
        for(int i=0;i<kInitColoring;i++)
            cutPlanes.modelAddCut(stablesInitColoring[i],1);
    }
    else {
        if(warmStartFromCol)
            cout<<"I could not start with a coloring from "<<colFile<<". ";
        cout<<"Generating simple input coloring with n="<<::n<<" colors, 1 color per vtx...";
        kInitColoring       = ::n;
        stablesInitColoring = new int*[kInitColoring];
        for(int i=0;i<(::n);i++){
            stablesInitColoring[i] = new int[::n];
            for(int j=0;j<(::n);j++)
                if(i==j)
                    stablesInitColoring[i][j]=1;
                else
                    stablesInitColoring[i][j]=0;
        }
        cout<<"Done\n";
    }

    //take initial dual values from initial coloring or a basic n-coloring
    double * dualsStart = new double[::n];
    for(int i=0;i<kInitColoring;i++){
        double stabSize=0;
        for(int j=0;j<n;j++)
            stabSize +=stablesInitColoring[i][j];
        for(int j=0;j<n;j++)
            if(stablesInitColoring[i][j])
                dualsStart[j] = 1.0/stabSize;
    }

    #ifndef NDEBUG
    double sumDuals=0;
    for(int i=0;i<n;i++)
        sumDuals += dualsStart[i];
    assert(abs(sumDuals-kInitColoring)<EPSILON);
    #endif
 
    cutPlanes.setPrimals(dualsStart,kInitColoring);

    int itersUsed;
    double CPUtimeUsed;                     //including time of cplex threads      
                                     
    time_t start = time(NULL);
    startTime = getCPUTime();
    if(cutPlanes.runCutPlanes(itersUsed, CPUtimeUsed)==EXIT_FAILURE)
        cout<<"\n\n ATTENTION: NOT enough time or iterations to optimize to the full!";
    double finalObj = cutPlanes.getObjVal();
    cout<<"\nFinal obj val="<<setprecision(10)<<finalObj<<" obtained after "<<itersUsed
        <<" iterations." <<" for nbColors="<<kInitColoring<<endl;
    cout<<"CPU Time:"<<CPUtimeUsed<<endl;
    cout<<"Real time:"<<time(NULL)-start<<endl;
    cout<<"First bound:"<<firstFarley<<endl;
    cout<<"Best bound:"<<bestLowerBnd<<endl;
    double*x = new double[n]; 

    cutPlanes.getPrimals(x); 
    for(int i=0;i<n;i++)
        if(x[i]>EPSILON)
            clog<<"x["<<i<<"]="<<x[i]<<"; ";
    clog<<endl; 

    if(!::useStdStabMaxSz)
        if(inter_with_std_stables==LST_IT_PROJ_STD_STB)
           oneAdditionalIter(argv[2]); 
}
